import json
import math
import random
import time
import os
import socket
import copy

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import KUBERNETES_SECRETS

CLIENT_REFRESH_INTERVAL_SECONDS = 300

class KubernetesJobException(MetaflowException):
    headline = "Kubernetes job error"

# Implements truncated exponential backoff from
# https://cloud.google.com/storage/docs/retry-strategy#exponential-backoff
def k8s_retry(deadline_seconds=60, max_backoff=32):
    def decorator(function):
        from functools import wraps

        @wraps(function)
        def wrapper(*args, **kwargs):
            from kubernetes import client

            deadline = time.time() + deadline_seconds
            retry_number = 0

            while True:
                try:
                    result = function(*args, **kwargs)
                    return result
                except client.rest.ApiException as e:
                    if e.status == 500:
                        current_t = time.time()
                        backoff_delay = min(
                            math.pow(2, retry_number) + random.random(), max_backoff
                        )
                        if current_t + backoff_delay < deadline:
                            time.sleep(backoff_delay)
                            retry_number += 1
                            continue  # retry again
                        else:
                            raise
                    else:
                        raise

        return wrapper

    return decorator


class KubernetesJob(object):
    def __init__(self, client, **kwargs):
        self._client = client
        self._kwargs = kwargs

    def create(self):
        # A discerning eye would notice and question the choice of using the
        # V1Job construct over the V1Pod construct given that we don't rely much
        # on any of the V1Job semantics. The major reasons at the moment are -
        #     1. It makes the Kubernetes UIs (Octant, Lens) a bit easier on
        #        the eyes, although even that can be questioned.
        #     2. AWS Step Functions, at the moment (Apr' 22) only supports
        #        executing Jobs and not Pods as part of it's publicly declared
        #        API. When we ship the AWS Step Functions integration with EKS,
        #        it will hopefully lessen our workload.
        #
        # Note: This implementation ensures that there is only one unique Pod
        # (unique UID) per Metaflow task attempt.
        client = self._client.get()

        # tmpfs variables
        use_tmpfs = self._kwargs["use_tmpfs"]
        tmpfs_size = self._kwargs["tmpfs_size"]
        tmpfs_enabled = use_tmpfs or (tmpfs_size and not use_tmpfs)

        jobset_name = "js-%s" % self._kwargs["attrs"]["metaflow.task_id"].split('-')[-1]
        main_job_name = "control"
        main_job_index = 0
        main_pod_index = 0
        subdomain = jobset_name
        master_port = int(self._kwargs['port']) if self._kwargs['port'] else None
        shared_memory = int(self._kwargs['shared_memory']) if self._kwargs['shared_memory'] else None

        passwordless_ssh = self._kwargs["attrs"]["requires_passwordless_ssh"]
        if passwordless_ssh:
            passwordless_ssh_service_name = subdomain
            passwordless_ssh_service_selector = {
                "passwordless-ssh-jobset": "true"
            }
        else:
            passwordless_ssh_service_name = None
            passwordless_ssh_service_selector = {}

        fqdn_suffix = "%s.svc.cluster.local" % self._kwargs["namespace"]
        jobset_main_addr = "%s-%s-%s-%s.%s.%s" % (
            jobset_name,
            main_job_name,
            main_job_index,
            main_pod_index,
            subdomain,
            fqdn_suffix,
        )

        def _install_jobset(
            repo_url="https://github.com/kubernetes-sigs/jobset",
            python_sdk_path="jobset/sdk/python",
        ):

            # TODO (Eddie): Remove this and suggest to user.

            import subprocess
            import tempfile
            import shutil
            import os

            with open(os.devnull, "wb") as devnull:
                cwd = os.getcwd()
                tmp_dir = tempfile.mkdtemp()
                os.chdir(tmp_dir)
                subprocess.check_call(
                    ["git", "clone", repo_url], stdout=devnull, stderr=subprocess.STDOUT
                )
                tmp_python_sdk_path = os.path.join(tmp_dir, python_sdk_path)
                os.chdir(tmp_python_sdk_path)
                subprocess.check_call(
                    ["pip", "install", "."], stdout=devnull, stderr=subprocess.STDOUT
                )
                os.chdir(cwd)
                shutil.rmtree(tmp_dir)

        def _get_passwordless_ssh_service():

            return client.V1Service(
                api_version="v1",
                kind="Service",
                metadata=client.V1ObjectMeta(
                    name=passwordless_ssh_service_name,
                    namespace=self._kwargs["namespace"]
                ),
                spec=client.V1ServiceSpec(
                    cluster_ip="None",
                    internal_traffic_policy="Cluster",
                    ip_families=["IPv4"],
                    ip_family_policy="SingleStack",
                    selector=passwordless_ssh_service_selector,
                    session_affinity="None",
                    type="ClusterIP",
                    ports=[
                        client.V1ServicePort(
                            name="control",
                            port=22,
                            protocol="TCP",
                            target_port=22
                        )
                    ]
                )
            )

        def _get_replicated_job(job_name, parallelism, command):
            return jobset.models.jobset_v1alpha2_replicated_job.JobsetV1alpha2ReplicatedJob(
                name=job_name,
                template=client.V1JobTemplateSpec(
                    metadata=client.V1ObjectMeta(
                        annotations=self._kwargs.get("annotations", {}),
                        labels=self._kwargs.get("labels", {}),
                        namespace=self._kwargs["namespace"],
                    ),
                    spec=client.V1JobSpec(
                        parallelism=parallelism,  # how many jobs can run at once
                        completions=parallelism,  # how many Pods the JobSet creates in total
                        backoff_limit=0,
                        ttl_seconds_after_finished=7
                        * 60
                        * 60
                        * 24,
                        template=client.V1PodTemplateSpec(
                            metadata=client.V1ObjectMeta(
                                annotations=self._kwargs.get("annotations", {}),
                                labels={
                                    **self._kwargs.get("labels", {}),
                                    **passwordless_ssh_service_selector, # TODO: necessary?
                                    # TODO: cluster-name, app.kubernetes.io/name necessary?
                                },
                                namespace=self._kwargs["namespace"],
                            ),
                            spec=client.V1PodSpec(
                                active_deadline_seconds=self._kwargs[
                                    "timeout_in_seconds"
                                ],
                                containers=[
                                    client.V1Container(
                                        command=command,
                                        ports=[client.V1ContainerPort(container_port=master_port)] if master_port and job_name=="control" else [],
                                        env=[
                                            client.V1EnvVar(name=k, value=str(v))
                                            for k, v in self._kwargs.get(
                                                "environment_variables", {}
                                            ).items()
                                        ]
                                        + [
                                            client.V1EnvVar(
                                                name=k,
                                                value_from=client.V1EnvVarSource(
                                                    field_ref=client.V1ObjectFieldSelector(
                                                        field_path=str(v)
                                                    )
                                                ),
                                            )
                                            for k, v in {
                                                "METAFLOW_KUBERNETES_POD_NAMESPACE": "metadata.namespace",
                                                "METAFLOW_KUBERNETES_POD_NAME": "metadata.name",
                                                "METAFLOW_KUBERNETES_POD_ID": "metadata.uid",
                                                "METAFLOW_KUBERNETES_SERVICE_ACCOUNT_NAME": "spec.serviceAccountName",
                                                "METAFLOW_KUBERNETES_NODE_IP": "status.hostIP",
                                            }.items()
                                        ]
                                        # Mimicking the AWS Batch Multinode env vars.
                                        + [
                                            client.V1EnvVar(
                                                name="MASTER_ADDR",
                                                value=jobset_main_addr,
                                            ),
                                            client.V1EnvVar(
                                                name="MASTER_PORT",
                                                value=str(master_port),
                                            ),
                                            client.V1EnvVar(
                                                name="RANK",
                                                value_from=client.V1EnvVarSource(
                                                    field_ref=client.V1ObjectFieldSelector(
                                                        field_path="metadata.annotations['batch.kubernetes.io/job-completion-index']"
                                                    )
                                                ),
                                            ),
                                            client.V1EnvVar(
                                                name="WORLD_SIZE",
                                                value=str(self._kwargs["num_parallel"]),
                                            ),
                                            client.V1EnvVar(
                                                name="PYTHONUNBUFFERED",
                                                value="0",
                                            ),
                                        ],
                                        env_from=[
                                            client.V1EnvFromSource(
                                                secret_ref=client.V1SecretEnvSource(
                                                    name=str(k),
                                                    # optional=True
                                                )
                                            )
                                            for k in list(
                                                self._kwargs.get("secrets", [])
                                            )
                                            + KUBERNETES_SECRETS.split(",")
                                            if k
                                        ],
                                        image=self._kwargs["image"],
                                        image_pull_policy=self._kwargs[
                                            "image_pull_policy"
                                        ],
                                        name=self._kwargs["step_name"].replace(
                                            "_", "-"
                                        ),
                                        resources=client.V1ResourceRequirements(
                                            requests={
                                                "cpu": str(self._kwargs["cpu"]),
                                                "memory": "%sM"
                                                % str(self._kwargs["memory"]),
                                                "ephemeral-storage": "%sM"
                                                % str(self._kwargs["disk"]),
                                            },
                                            limits={
                                                "%s.com/gpu".lower()
                                                % self._kwargs["gpu_vendor"]: str(
                                                    self._kwargs["gpu"]
                                                )
                                                for k in [0]
                                                # Don't set GPU limits if gpu isn't specified.
                                                if self._kwargs["gpu"] is not None
                                            },
                                        ),
                                        volume_mounts=(
                                            [
                                                client.V1VolumeMount(
                                                    mount_path=self._kwargs.get(
                                                        "tmpfs_path"
                                                    ),
                                                    name="tmpfs-ephemeral-volume",
                                                )
                                            ]
                                            if tmpfs_enabled
                                            else []
                                        )
                                        + (
                                            [
                                                client.V1VolumeMount(
                                                    mount_path="/dev/shm",
                                                    name="dhsm"
                                                )
                                            ]
                                            if shared_memory else []
                                        )
                                        + (
                                            [
                                                client.V1VolumeMount(
                                                    mount_path=path, name=claim
                                                )
                                                for claim, path in self._kwargs[
                                                    "persistent_volume_claims"
                                                ].items()
                                            ]
                                            if self._kwargs["persistent_volume_claims"]
                                            is not None
                                            else []
                                        ),
                                    )
                                ],
                                node_selector=self._kwargs.get("node_selector"),
                                restart_policy="Never",

                                set_hostname_as_fqdn=True, # configure pod hostname as pod's FQDN
                                share_process_namespace=False, # default
                                subdomain=subdomain, # FQDN = <hostname>.<subdomain>.<pod namespace>.svc.<cluster domain>

                                service_account_name=self._kwargs["service_account"],
                                termination_grace_period_seconds=0,
                                tolerations=[
                                    client.V1Toleration(**toleration)
                                    for toleration in self._kwargs.get("tolerations")
                                    or []
                                ],
                                volumes=(
                                    [
                                        client.V1Volume(
                                            name="tmpfs-ephemeral-volume",
                                            empty_dir=client.V1EmptyDirVolumeSource(
                                                medium="Memory",
                                                size_limit="{}Mi".format(tmpfs_size),
                                            ),
                                        )
                                    ]
                                    if tmpfs_enabled
                                    else []
                                )
                                + (
                                    [
                                        client.V1Volume(
                                            name="dhsm",
                                            empty_dir=client.V1EmptyDirVolumeSource(
                                                medium="Memory",
                                                size_limit="{}Mi".format(shared_memory),
                                            )
                                        )
                                    ]
                                    if shared_memory else []
                                )
                                + (
                                    [
                                        client.V1Volume(
                                            name=claim,
                                            persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(
                                                claim_name=claim
                                            ),
                                        )
                                        for claim in self._kwargs[
                                            "persistent_volume_claims"
                                        ].keys()
                                    ]
                                    if self._kwargs["persistent_volume_claims"]
                                    is not None
                                    else []
                                ),
                            ),
                        ),
                    ),
                ),
            )

        if "num_parallel" in self._kwargs and self._kwargs["num_parallel"] >= 1:

            try:
                import jobset
            except ImportError:
                _install_jobset()
                import jobset

            main_commands = copy.copy(self._kwargs["command"])
            main_commands[-1] = main_commands[-1].replace(
                "[multinode-args]", "--split-index 0"
            )

            task_id = self._kwargs["attrs"]["metaflow.task_id"]
            secondary_commands = copy.copy(self._kwargs["command"])
            # RANK needs +1 because control node is not in the worker index group, yet we want global nodes.
            # Technically, control and worker could be same replicated job type, but cleaner to separate for future use cases.
            secondary_commands[-1] = secondary_commands[-1].replace(
                "[multinode-args]", "--split-index `expr $RANK + 1`"
            )
            secondary_commands[-1] = secondary_commands[-1].replace(
                "ubf_control", "ubf_task"
            )
            secondary_commands[-1] = secondary_commands[-1].replace(
                task_id,
                task_id.replace("control-", "") + "-node-`expr $RANK + 1`",
            )

            if passwordless_ssh:
                if not os.path.exists("/usr/sbin/sshd"):
                    raise KubernetesJobException(
                        "This @parallel decorator requires sshd to be installed in the container image."
                        "Please install OpenSSH."
                    )

                # run sshd in background
                main_commands[-1] = "/usr/sbin/sshd -D & %s" % main_commands[-1]
                secondary_commands[-1] = "/usr/sbin/sshd -D & %s" % secondary_commands[-1]

            self._jobset = jobset.models.jobset_v1alpha2_job_set.JobsetV1alpha2JobSet(
                api_version="jobset.x-k8s.io/v1alpha2",
                kind="JobSet",
                metadata=client.V1ObjectMeta(
                    annotations=self._kwargs.get("annotations", {}),
                    labels=self._kwargs.get("labels", {}),
                    name=jobset_name,
                    namespace=self._kwargs["namespace"],
                ),
                spec=jobset.models.jobset_v1alpha2_job_set_spec.JobsetV1alpha2JobSetSpec(
                    network=jobset.models.jobset_v1alpha2_network.JobsetV1alpha2Network(
                        enable_dns_hostnames=True if not self._kwargs['attrs']['requires_passwordless_ssh'] else False,
                        subdomain=subdomain
                    ),
                    replicated_jobs=[
                        _get_replicated_job("control", 1, main_commands),
                        _get_replicated_job(
                            "worker",
                            self._kwargs["num_parallel"] - 1,
                            secondary_commands,
                        ),
                    ],
                ),
            )
            self._passwordless_ssh_service = _get_passwordless_ssh_service()
        else:
            self._job = client.V1Job(
                api_version="batch/v1",
                kind="Job",
                metadata=client.V1ObjectMeta(
                    # Annotations are for humans
                    annotations=self._kwargs.get("annotations", {}),
                    # While labels are for Kubernetes
                    labels=self._kwargs.get("labels", {}),
                    generate_name=self._kwargs["generate_name"],
                    namespace=self._kwargs["namespace"],  # Defaults to `default`
                ),
                spec=client.V1JobSpec(
                    # Retries are handled by Metaflow when it is responsible for
                    # executing the flow. The responsibility is moved to Kubernetes
                    # when Argo Workflows is responsible for the execution.
                    backoff_limit=self._kwargs.get("retries", 0),
                    completions=1,  # A single non-indexed pod job
                    ttl_seconds_after_finished=7
                    * 60
                    * 60  # Remove job after a week. TODO: Make this configurable
                    * 24,
                    template=client.V1PodTemplateSpec(
                        metadata=client.V1ObjectMeta(
                            annotations=self._kwargs.get("annotations", {}),
                            labels=self._kwargs.get("labels", {}),
                            namespace=self._kwargs["namespace"],
                        ),
                        spec=client.V1PodSpec(
                            # Timeout is set on the pod and not the job (important!)
                            active_deadline_seconds=self._kwargs["timeout_in_seconds"],
                            # TODO (savin): Enable affinities for GPU scheduling.
                            # affinity=?,
                            containers=[
                                client.V1Container(
                                    command=self._kwargs["command"],
                                    env=[
                                        client.V1EnvVar(name=k, value=str(v))
                                        for k, v in self._kwargs.get(
                                            "environment_variables", {}
                                        ).items()
                                    ]
                                    # And some downward API magic. Add (key, value)
                                    # pairs below to make pod metadata available
                                    # within Kubernetes container.
                                    + [
                                        client.V1EnvVar(
                                            name=k,
                                            value_from=client.V1EnvVarSource(
                                                field_ref=client.V1ObjectFieldSelector(
                                                    field_path=str(v)
                                                )
                                            ),
                                        )
                                        for k, v in {
                                            "METAFLOW_KUBERNETES_POD_NAMESPACE": "metadata.namespace",
                                            "METAFLOW_KUBERNETES_POD_NAME": "metadata.name",
                                            "METAFLOW_KUBERNETES_POD_ID": "metadata.uid",
                                            "METAFLOW_KUBERNETES_SERVICE_ACCOUNT_NAME": "spec.serviceAccountName",
                                            "METAFLOW_KUBERNETES_NODE_IP": "status.hostIP",
                                        }.items()
                                    ],
                                    env_from=[
                                        client.V1EnvFromSource(
                                            secret_ref=client.V1SecretEnvSource(
                                                name=str(k),
                                                # optional=True
                                            )
                                        )
                                        for k in list(self._kwargs.get("secrets", []))
                                        + KUBERNETES_SECRETS.split(",")
                                        if k
                                    ],
                                    image=self._kwargs["image"],
                                    image_pull_policy=self._kwargs["image_pull_policy"],
                                    name=self._kwargs["step_name"].replace("_", "-"),
                                    resources=client.V1ResourceRequirements(
                                        requests={
                                            "cpu": str(self._kwargs["cpu"]),
                                            "memory": "%sM"
                                            % str(self._kwargs["memory"]),
                                            "ephemeral-storage": "%sM"
                                            % str(self._kwargs["disk"]),
                                        },
                                        limits={
                                            "%s.com/gpu".lower()
                                            % self._kwargs["gpu_vendor"]: str(
                                                self._kwargs["gpu"]
                                            )
                                            for k in [0]
                                            # Don't set GPU limits if gpu isn't specified.
                                            if self._kwargs["gpu"] is not None
                                        },
                                    ),
                                    volume_mounts=(
                                        [
                                            client.V1VolumeMount(
                                                mount_path=self._kwargs.get(
                                                    "tmpfs_path"
                                                ),
                                                name="tmpfs-ephemeral-volume",
                                            )
                                        ]
                                        if tmpfs_enabled
                                        else []
                                    )
                                    + (
                                        [
                                            client.V1VolumeMount(
                                                mount_path=path, name=claim
                                            )
                                            for claim, path in self._kwargs[
                                                "persistent_volume_claims"
                                            ].items()
                                        ]
                                        if self._kwargs["persistent_volume_claims"]
                                        is not None
                                        else []
                                    ),
                                )
                            ],
                            node_selector=self._kwargs.get("node_selector"),
                            # TODO (savin): Support image_pull_secrets
                            # image_pull_secrets=?,
                            # TODO (savin): Support preemption policies
                            # preemption_policy=?,
                            #
                            # A Container in a Pod may fail for a number of
                            # reasons, such as because the process in it exited
                            # with a non-zero exit code, or the Container was
                            # killed due to OOM etc. If this happens, fail the pod
                            # and let Metaflow handle the retries.
                            restart_policy="Never",
                            service_account_name=self._kwargs["service_account"],
                            # Terminate the container immediately on SIGTERM
                            termination_grace_period_seconds=0,
                            tolerations=[
                                client.V1Toleration(**toleration)
                                for toleration in self._kwargs.get("tolerations") or []
                            ],
                            volumes=(
                                [
                                    client.V1Volume(
                                        name="tmpfs-ephemeral-volume",
                                        empty_dir=client.V1EmptyDirVolumeSource(
                                            medium="Memory",
                                            # Add default unit as ours differs from Kubernetes default.
                                            size_limit="{}Mi".format(tmpfs_size),
                                        ),
                                    )
                                ]
                                if tmpfs_enabled
                                else []
                            )
                            + (
                                [
                                    client.V1Volume(
                                        name=claim,
                                        persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(
                                            claim_name=claim
                                        ),
                                    )
                                    for claim in self._kwargs[
                                        "persistent_volume_claims"
                                    ].keys()
                                ]
                                if self._kwargs["persistent_volume_claims"] is not None
                                else []
                            ),
                            # TODO (savin): Set termination_message_policy
                        ),
                    ),
                ),
            )
        return self

    def execute(self):
        client = self._client.get()
        try:
            # TODO: Make job submission back-pressure aware. Currently
            #       there doesn't seem to be a kubernetes-native way to
            #       achieve the guarantees that we are seeking.
            #       https://github.com/kubernetes/enhancements/issues/1040
            #       Hopefully, we will be able to get creative with kube-batch

            if "num_parallel" in self._kwargs and self._kwargs["num_parallel"] >= 1:
                # TODO (Eddie): this is kinda gross. fix it.
                if self._kwargs["attrs"]["requires_passwordless_ssh"]:
                    api_instance = client.CoreV1Api()
                    api_response = api_instance.create_namespaced_service(namespace=self._kwargs['namespace'], body=self._passwordless_ssh_service)

                with client.ApiClient() as api_client:
                    api_instance = client.CustomObjectsApi(api_client)

                response = api_instance.create_namespaced_custom_object(
                    body=self._jobset,
                    group="jobset.x-k8s.io",
                    version="v1alpha2",
                    namespace=self._kwargs["namespace"],
                    plural="jobsets",
                )

                # HACK: Give K8s some time to actually create the job
                time.sleep(10)

                # TODO (Eddie): Remove hack and make RunningJobSet.
                # There are many jobs running that should be monitored.
                job_name = "%s-control-0" % response["metadata"]["name"]
                fake_id = 123
                return RunningJob(
                    client=self._client,
                    name=job_name,
                    uid=fake_id,
                    namespace=response["metadata"]["namespace"],
                )

            else:
                response = (
                    client.BatchV1Api()
                    .create_namespaced_job(
                        body=self._job, namespace=self._kwargs["namespace"]
                    )
                    .to_dict()
                )
                return RunningJob(
                    client=self._client,
                    name=response["metadata"]["name"],
                    uid=response["metadata"]["uid"],
                    namespace=response["metadata"]["namespace"],
                )

        except client.rest.ApiException as e:
            raise KubernetesJobException(
                "Unable to launch Kubernetes job.\n %s"
                % (json.loads(e.body)["message"] if e.body is not None else e.reason)
            )

    def step_name(self, step_name):
        self._kwargs["step_name"] = step_name
        return self

    def namespace(self, namespace):
        self._kwargs["namespace"] = namespace
        return self

    def name(self, name):
        self._kwargs["name"] = name
        return self

    def command(self, command):
        self._kwargs["command"] = command
        return self

    def image(self, image):
        self._kwargs["image"] = image
        return self

    def cpu(self, cpu):
        self._kwargs["cpu"] = cpu
        return self

    def memory(self, mem):
        self._kwargs["memory"] = mem
        return self

    def environment_variable(self, name, value):
        # Never set to None
        if value is None:
            return self
        self._kwargs["environment_variables"] = dict(
            self._kwargs.get("environment_variables", {}), **{name: value}
        )
        return self

    def label(self, name, value):
        self._kwargs["labels"] = dict(self._kwargs.get("labels", {}), **{name: value})
        return self

    def annotation(self, name, value):
        self._kwargs["annotations"] = dict(
            self._kwargs.get("annotations", {}), **{name: value}
        )
        return self


class RunningJob(object):
    # State Machine implementation for the lifecycle behavior documented in
    # https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/
    #
    # This object encapsulates *both* V1Job and V1Pod. It simplifies the status
    # to "running" and "done" (failed/succeeded) state. Note that V1Job and V1Pod
    # status fields are not guaranteed to be always in sync due to the way job
    # controller works.

    # To ascertain the status of V1Job, we peer into the lifecycle status of
    # the pod it is responsible for executing. Unfortunately, the `phase`
    # attributes (pending, running, succeeded, failed etc.) only provide
    # partial answers and the official API conventions guide suggests that
    # it may soon be deprecated (however, not anytime soon - see
    # https://github.com/kubernetes/kubernetes/issues/7856). `conditions` otoh
    # provide a deeper understanding about the state of the pod; however
    # conditions are not state machines and can be oscillating - from the
    # official API conventions guide:
    #     In general, condition values may change back and forth, but some
    #     condition transitions may be monotonic, depending on the resource and
    #     condition type. However, conditions are observations and not,
    #     themselves, state machines, nor do we define comprehensive state
    #     machines for objects, nor behaviors associated with state
    #     transitions. The system is level-based rather than edge-triggered,
    #     and should assume an Open World.
    # As a follow-up, we can synthesize our notion of "phase" state
    # machine from `conditions`, since Kubernetes won't do it for us (for
    # many good reasons).
    #
    # `conditions` can be of the following types -
    #    1. (kubelet) Initialized (always True since we don't rely on init
    #       containers)
    #    2. (kubelet) ContainersReady
    #    3. (kubelet) Ready (same as ContainersReady since we don't use
    #       ReadinessGates -
    #       https://github.com/kubernetes/kubernetes/blob/master/pkg/kubelet/status/generate.go)
    #    4. (kube-scheduler) PodScheduled
    #       (https://github.com/kubernetes/kubernetes/blob/master/pkg/scheduler/scheduler.go)

    def __init__(self, client, name, uid, namespace):
        self._client = client
        self._name = name
        self._pod_name = None
        self._id = uid
        self._namespace = namespace

        self._job = self._fetch_job()
        self._pod = self._fetch_pod()

        import atexit

        def best_effort_kill():
            try:
                self.kill()
            except:
                pass

        atexit.register(best_effort_kill)

    def __repr__(self):
        return "{}('{}/{}')".format(
            self.__class__.__name__, self._namespace, self._name
        )

    @k8s_retry()
    def _fetch_job(self):
        client = self._client.get()
        try:
            return (
                client.BatchV1Api()
                .read_namespaced_job(name=self._name, namespace=self._namespace)
                .to_dict()
            )
        except client.rest.ApiException as e:
            if e.status == 404:
                raise KubernetesJobException(
                    "Unable to locate Kubernetes batch/v1 job %s" % self._name
                )
            raise

    @k8s_retry()
    def _fetch_pod(self):
        # Fetch pod metadata.
        client = self._client.get()
        pods = (
            client.CoreV1Api()
            .list_namespaced_pod(
                namespace=self._namespace,
                label_selector="job-name={}".format(self._name),
            )
            .to_dict()["items"]
        )
        if pods:
            return pods[0]
        return {}

    def kill(self):
        # Terminating a Kubernetes job is a bit tricky. Issuing a
        # `BatchV1Api.delete_namespaced_job` will also remove all traces of the
        # job object from the Kubernetes API server which may not be desirable.
        # This forces us to be a bit creative in terms of how we handle kill:
        #
        # 1. If the container is alive and kicking inside the pod, we simply
        #    attach ourselves to the container and issue a kill signal. The
        #    way we have initialized the Job ensures that the job will cleanly
        #    terminate.
        # 2. In scenarios where either the pod (unschedulable etc.) or the
        #    container (ImagePullError etc.) hasn't come up yet, we become a
        #    bit creative by patching the job parallelism to 0. This ensures
        #    that the underlying node's resources are made available to
        #    kube-scheduler again. The downside is that the Job wouldn't mark
        #    itself as done and the pod metadata disappears from the API
        #    server. There is an open issue in the Kubernetes GH to provide
        #    better support for job terminations -
        #    https://github.com/kubernetes/enhancements/issues/2232
        # 3. If the pod object hasn't shown up yet, we set the parallelism to 0
        #    to preempt it.
        client = self._client.get()
        if not self.is_done:
            if self.is_running:
                # Case 1.
                from kubernetes.stream import stream

                api_instance = client.CoreV1Api
                try:
                    # TODO: stream opens a web-socket connection. It may
                    #       not be desirable to open multiple web-socket
                    #       connections frivolously (think killing a
                    #       workflow during a for-each step).
                    stream(
                        api_instance().connect_get_namespaced_pod_exec,
                        name=self._pod["metadata"]["name"],
                        namespace=self._namespace,
                        command=[
                            "/bin/sh",
                            "-c",
                            "/sbin/killall5",
                        ],
                        stderr=True,
                        stdin=False,
                        stdout=True,
                        tty=False,
                    )
                except:
                    # Best effort. It's likely that this API call could be
                    # blocked for the user.
                    # --------------------------------------------------------
                    # We try patching Job parallelism anyway. Stopping any runaway
                    # jobs (and their pods) is secondary to correctly showing
                    # "Killed" status on the Kubernetes pod.
                    #
                    # This has the effect of pausing the job.
                    try:
                        client.BatchV1Api().patch_namespaced_job(
                            name=self._name,
                            namespace=self._namespace,
                            field_manager="metaflow",
                            body={"spec": {"parallelism": 0}},
                        )
                    except:
                        # Best effort.
                        pass
                        # raise
            else:
                # Case 2.
                # This has the effect of pausing the job.
                try:
                    client.BatchV1Api().patch_namespaced_job(
                        name=self._name,
                        namespace=self._namespace,
                        field_manager="metaflow",
                        body={"spec": {"parallelism": 0}},
                    )
                except:
                    # Best effort.
                    pass
                    # raise
        return self

    @property
    def id(self):
        if self._pod_name:
            return "pod %s" % self._pod_name
        if self._pod:
            self._pod_name = self._pod["metadata"]["name"]
            return self.id
        return "job %s" % self._name

    @property
    def is_done(self):
        # Check if the container is done. As a side effect, also refreshes self._job and
        # self._pod with the latest state
        def done():
            # Either the container succeeds or fails naturally or else we may have
            # forced the pod termination causing the job to still be in an
            # active state but for all intents and purposes dead to us.
            return (
                bool(self._job["status"].get("succeeded"))
                or bool(self._job["status"].get("failed"))
                or self._are_pod_containers_done
                or (self._job["spec"]["parallelism"] == 0)
            )

        if not done():
            # If not done, fetch newer status
            self._job = self._fetch_job()
            self._pod = self._fetch_pod()
        return done()

    @property
    def status(self):
        if not self.is_done:
            if bool(self._job["status"].get("active")):
                if self._pod:
                    msg = (
                        "Pod is %s"
                        % self._pod.get("status", {})
                        .get("phase", "uninitialized")
                        .lower()
                    )
                    # TODO (savin): parse Pod conditions
                    container_status = (
                        self._pod["status"].get("container_statuses") or [None]
                    )[0]
                    if container_status:
                        # We have a single container inside the pod
                        status = {"status": "waiting"}
                        for k, v in container_status["state"].items():
                            if v is not None:
                                status["status"] = k
                                status.update(v)
                        msg += ", Container is %s" % status["status"].lower()
                        reason = ""
                        if status.get("reason"):
                            pass
                            reason = status["reason"]
                        if status.get("message"):
                            reason += " - %s" % status["message"]
                        if reason:
                            msg += " - %s" % reason
                    return msg
                return "Job is active"
            return "Job status is unknown"
        return "Job is done"

    @property
    def has_succeeded(self):
        # The tasks container is in a terminal state and the status is marked as succeeded
        return self.is_done and self._have_containers_succeeded

    @property
    def has_failed(self):
        # Either the container is marked as failed or the Job is not allowed to
        # any more pods
        retval = self.is_done and (
            bool(self._job["status"].get("failed"))
            or self._has_any_container_failed
            or (self._job["spec"]["parallelism"] == 0)
        )
        return retval

    @property
    def _have_containers_succeeded(self):
        container_statuses = self._pod.get("status", {}).get("container_statuses", [])
        if not container_statuses:
            return False

        for cstatus in container_statuses:
            # If the terminated field is not set, the pod is still running.
            terminated = cstatus.get("state", {}).get("terminated", {})
            if not terminated:
                return False

            # If the terminated field is set but the `finished_at` field is not set,
            # the pod is still considered as running.
            if not terminated.get("finished_at"):
                return False

            # If finished_at is set AND reason is Completed
            if terminated.get("reason", "").lower() != "completed":
                return False

        return True

    @property
    def _has_any_container_failed(self):
        container_statuses = self._pod.get("status", {}).get("container_statuses", [])
        if not container_statuses:
            return False

        for cstatus in container_statuses:
            # If the terminated field is not set, the pod is still running. Too early
            # to determine if any container failed.
            terminated = cstatus.get("state", {}).get("terminated", {})
            if not terminated:
                return False

            # If the terminated field is set but the `finished_at` field is not set,
            # the pod is still considered as running. Too early to determine if any
            # container failed.
            if not terminated.get("finished_at"):
                return False

            # If finished_at is set AND reason is Error, it means that the
            # container failed.
            if terminated.get("reason", "").lower() == "error":
                return True

        # If none of the containers are marked as failed, the pod is not
        # considered failed.
        return False

    @property
    def _are_pod_containers_done(self):
        # All containers in the pod have a containerStatus that has a
        # finishedAt set.
        container_statuses = self._pod.get("status", {}).get("container_statuses", [])
        if not container_statuses:
            return False

        for cstatus in container_statuses:
            # If the terminated field is not set, the pod is still running. Too early
            # to determine if any container failed.
            terminated = cstatus.get("state", {}).get("terminated", {})
            if not terminated:
                return False

            # If the terminated field is set but the `finished_at` field is not set,
            # the pod is still considered as running.
            if not terminated.get("finished_at"):
                return False

        # If we got until here, the containers were marked terminated and their
        # finishedAt was set.
        return True

    @property
    def is_running(self):
        # Returns true if the container is running.
        if self.is_done:
            return False

        return not self._are_pod_containers_done

    @property
    def is_waiting(self):
        return not self.is_done and not self.is_running

    @property
    def reason(self):
        if self.is_done:
            if self.has_succeeded:
                return 0, None
            # Best effort since Pod object can disappear on us at anytime
            else:
                if self._pod.get("status", {}).get("phase") not in (
                    "Succeeded",
                    "Failed",
                ):
                    # If pod status is dirty, check for newer status
                    self._pod = self._fetch_pod()
                if self._pod:
                    if self._pod.get("status", {}).get("container_statuses") is None:
                        # We're done, but no container_statuses is set
                        # This can happen when the pod is evicted
                        return None, ": ".join(
                            filter(
                                None,
                                [
                                    self._pod.get("status", {}).get("reason"),
                                    self._pod.get("status", {}).get("message"),
                                ],
                            )
                        )

                    for k, v in (
                        self._pod.get("status", {})
                        .get("container_statuses", [{}])[0]
                        .get("state", {})
                        .items()
                    ):
                        if v is not None:
                            return v.get("exit_code"), ": ".join(
                                filter(
                                    None,
                                    [v.get("reason"), v.get("message")],
                                )
                            )
        return None, None



import base64
import json
import os
from .card import MetaflowCard, MetaflowCardComponent
from .convert_to_native_type import TaskToDict
import uuid

ABS_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
RENDER_TEMPLATE_PATH = os.path.join(ABS_DIR_PATH, "base.html")
JS_PATH = os.path.join(ABS_DIR_PATH, "main.js")
CSS_PATH = os.path.join(ABS_DIR_PATH, "bundle.css")


def transform_flow_graph(step_info):
    def node_to_type(node_type):
        if node_type in ["linear", "start", "end", "join"]:
            return node_type
        elif node_type == "split":
            return "split"
        elif node_type == "split-parallel" or node_type == "split-foreach":
            return "foreach"
        return "unknown"  # Should never happen

    graph_dict = {}
    for stepname in step_info:
        graph_dict[stepname] = {
            "type": node_to_type(step_info[stepname]["type"]),
            "box_next": step_info[stepname]["type"] not in ("linear", "join"),
            "box_ends": None
            if "matching_join" not in step_info[stepname]
            else step_info[stepname]["matching_join"],
            "next": step_info[stepname]["next"],
            "doc": step_info[stepname]["doc"],
        }
    return graph_dict


def read_file(path):
    with open(path, "r") as f:
        return f.read()


class DefaultComponent(MetaflowCardComponent):
    """
    The `DefaultCard` and the `BlankCard` use a JS framework that build the HTML dynamically from JSON.
    The `DefaultComponent` is the base component that helps build the JSON when `render` is called.

    The underlying JS framework consists of various types of objects.
    These can be found in: "metaflow/plugins/cards/ui/types.ts".
    The `type` attribute in a `DefaultComponent` corresponds to the type of component in the Javascript framework.
    """

    type = None

    def __init__(self, title=None, subtitle=None):
        self._title = title
        self._subtitle = subtitle

    def render(self):
        datadict = dict(
            type=self.type,
        )
        if self._title is not None:
            datadict["title"] = self._title
        if self._subtitle is not None:
            datadict["subtitle"] = self._subtitle
        return datadict


class TitleComponent(MetaflowCardComponent):
    type = "title"

    def __init__(self, text=None):
        self._text = text

    def render(self):
        return dict(type=self.type, text=str(self._text))


class SubTitleComponent(MetaflowCardComponent):
    type = "subtitle"

    def __init__(self, text=None):
        self._text = text

    def render(self):
        return dict(type=self.type, text=str(self._text))


class SectionComponent(DefaultComponent):
    type = "section"

    def __init__(self, title=None, subtitle=None, columns=None, contents=[]):
        super().__init__(title=title, subtitle=subtitle)
        # Contents are expected to be list of dictionaries.
        self._contents = contents
        self._columns = columns

    @classmethod
    def render_subcomponents(
        cls, component_array, additional_allowed_types=[str, dict], allow_unknowns=False
    ):
        contents = []
        for content in component_array:
            # Render objects of `MetaflowCardComponent` type
            if issubclass(type(content), MetaflowCardComponent):
                rendered_content = content.render()
                if type(rendered_content) == str or type(rendered_content) == dict:
                    contents.append(rendered_content)
                else:
                    contents.append(
                        SerializationErrorComponent(
                            content.__class__.__name__,
                            "Component render didn't return a string or dict",
                        ).render()
                    )
            # Objects of allowed types should be present.
            elif type(content) in additional_allowed_types:
                contents.append(content)
            elif allow_unknowns:
                contents.append("<object>")

        return contents

    def render(self):
        datadict = super().render()
        contents = self.render_subcomponents(self._contents)
        datadict["contents"] = contents
        if self._columns is not None:
            datadict["columns"] = self._columns
        return datadict


class ImageComponent(DefaultComponent):
    type = "image"

    def __init__(self, src=None, label=None, title=None, subtitle=None):
        super().__init__(title=title, subtitle=subtitle)
        self._src = src
        self._label = label

    def render(self):
        datadict = super().render()
        img_dict = dict(
            src=self._src,
            label=self._label,
        )
        datadict.update(img_dict)
        if self.component_id is not None:
            datadict["id"] = self.component_id
        return datadict


class TableComponent(DefaultComponent):
    type = "table"

    def __init__(
        self, title=None, subtitle=None, headers=[], data=[[]], vertical=False
    ):
        super().__init__(title=title, subtitle=subtitle)
        self._headers = []
        self._data = [[]]
        self._vertical = vertical

        if self._validate_header_type(headers):
            self._headers = headers
        if self._validate_row_type(data):
            self._data = data

    @classmethod
    def validate(cls, headers, data):
        return (cls._validate_header_type(headers), cls._validate_row_type(data))

    @staticmethod
    def _validate_header_type(data):
        if type(data) != list:
            return False
        return True

    @staticmethod
    def _validate_row_type(data):
        if type(data) != list:
            return False
        try:
            if type(data[0]) != list:
                return False
        except IndexError:
            return False
        except TypeError:
            return False

        return True

    def render(self):
        datadict = super().render()
        datadict["columns"] = self._headers
        datadict["data"] = self._data
        datadict["vertical"] = self._vertical
        if self.component_id is not None:
            datadict["id"] = self.component_id
        return datadict


class DagComponent(DefaultComponent):
    type = "dag"

    def __init__(self, title=None, subtitle=None, data={}):
        super().__init__(title=title, subtitle=subtitle)
        self._data = data

    def render(self):
        datadict = super().render()
        datadict["data"] = self._data
        return datadict


class TextComponent(DefaultComponent):
    type = "text"

    def __init__(self, text=None):
        super().__init__(title=None, subtitle=None)
        self._text = text

    def render(self):
        datadict = super().render()
        datadict["text"] = self._text
        return datadict


class LogComponent(DefaultComponent):
    type = "log"

    def __init__(self, data=None):
        super().__init__(title=None, subtitle=None)
        self._data = data

    def render(self):
        datadict = super().render()
        datadict["data"] = self._data
        return datadict


class HTMLComponent(DefaultComponent):
    type = "html"

    def __init__(self, data=None):
        super().__init__(title=None, subtitle=None)
        self._data = data

    def render(self):
        datadict = super().render()
        datadict["data"] = self._data
        return datadict


class PageComponent(DefaultComponent):
    type = "page"

    def __init__(self, title=None, subtitle=None, contents=[]):
        super().__init__(title=title, subtitle=subtitle)
        self._contents = contents

    def render(self):
        datadict = super().render()
        contents = []
        for content in self._contents:
            if issubclass(type(content), MetaflowCardComponent):
                contents.append(content.render())
            else:
                contents.append(content)
        datadict["contents"] = contents
        return datadict


class ErrorComponent(MetaflowCardComponent):
    def __init__(self, headline, error_message):
        self._headline = headline
        self._error_message = error_message

    def render(self):
        return LogComponent(
            data="%s\n\n%s" % (self._headline, self._error_message)
        ).render()


class SerializationErrorComponent(ErrorComponent):
    def __init__(self, component_name, error_message):
        headline = "Render failed of component named `%s`" % component_name
        super().__init__(headline, error_message)


class ArtifactsComponent(DefaultComponent):
    type = "artifacts"

    def __init__(self, title=None, subtitle=None, data={}):
        super().__init__(title=title, subtitle=subtitle)
        self._data = data

    def render(self):
        datadict = super().render()
        datadict["data"] = self._data
        if self.component_id is not None:
            datadict["id"] = self.component_id
        return datadict


class MarkdownComponent(DefaultComponent):
    type = "markdown"

    def __init__(self, text=None):
        super().__init__(title=None, subtitle=None)
        self._text = text

    def render(self):
        datadict = super().render()
        _text = self._text
        # When we have a markdown text that doesn't start with a `#`,
        # we need to add a newline to make sure that the markdown
        # is rendered correctly. Otherwise `svelte-markdown` will render
        # the empty `<p>` tags during re-renders on card data updates.
        if self._text is not None and not self._text.startswith("#"):
            _text = "\n" + _text
        datadict["source"] = _text
        if self.component_id is not None:
            datadict["id"] = self.component_id
        return datadict


class TaskInfoComponent(MetaflowCardComponent):
    """
    Properties
        page_content : a list of MetaflowCardComponents going as task info
        final_component: the dictionary returned by the `render` function of this class.
    """

    def __init__(
        self,
        task,
        page_title="Task Info",
        only_repr=True,
        graph=None,
        components=[],
        runtime=False,
    ):
        self._task = task
        self._only_repr = only_repr
        self._graph = graph
        self._components = components
        self._page_title = page_title
        self.final_component = None
        self.page_component = None
        self.runtime = runtime

    def render(self):
        """

        Returns:
            a dictionary of form:
                dict(metadata = {},components= [])
        """
        task_data_dict = TaskToDict(only_repr=self._only_repr)(
            self._task, graph=self._graph
        )
        # ignore the name as an artifact
        if "name" in task_data_dict["data"]:
            del task_data_dict["data"]["name"]

        _metadata = dict(version=1, template="defaultCardTemplate")
        # try to parse out metaflow version from tags, but let it go if unset
        # e.g. if a run came from a local, un-versioned metaflow codebase
        try:
            _metadata["metaflow_version"] = [
                t for t in self._task.parent.parent.tags if "metaflow_version" in t
            ][0].split("metaflow_version:")[1]
        except Exception:
            pass

        final_component_dict = dict(
            metadata=_metadata,
            components=[],
        )

        metadata = [
            "stderr",
            "stdout",
            "created_at",
            "finished_at",
            "pathspec",
        ]
        tags = self._task.parent.parent.tags
        user_info = [t for t in tags if t.startswith("user:")]
        task_metadata_dict = {
            "Task Created On": task_data_dict["created_at"],
            "Task Finished On": task_data_dict["finished_at"],
            # Remove Microseconds from timedelta
            "Tags": ", ".join(tags),
        }
        if not self.runtime:
            task_metadata_dict["Task Duration"] = str(
                self._task.finished_at - self._task.created_at
            ).split(".")[0]
        if len(user_info) > 0:
            task_metadata_dict["User"] = user_info[0].split("user:")[1]

        for m in metadata:
            final_component_dict["metadata"][m] = task_data_dict[m]

        metadata_table = SectionComponent(
            title="Task Metadata",
            contents=[
                TableComponent(
                    headers=list(task_metadata_dict.keys()),
                    data=[list(task_metadata_dict.values())],
                    vertical=True,
                )
            ],
        )

        img_components = []
        for img_name in task_data_dict["images"]:
            img_components.append(
                ImageComponent(
                    src=task_data_dict["images"][img_name], label=img_name
                ).render()
            )
        table_comps = []
        for tabname in task_data_dict["tables"]:
            tab_dict = task_data_dict["tables"][tabname]
            tab_title = "Artifact Name: %s" % tabname
            sec_tab_comp = [
                TableComponent(headers=tab_dict["headers"], data=tab_dict["data"])
            ]
            post_table_md = None

            if tab_dict["truncated"]:
                tab_title = "Artifact Name: %s (%d columns and %d rows)" % (
                    tabname,
                    tab_dict["full_size"][1],
                    tab_dict["full_size"][0],
                )
                post_table_md = MarkdownComponent(
                    "_Truncated - %d rows not shown_"
                    % ((tab_dict["full_size"][0] - len(tab_dict["data"])))
                )

            if post_table_md:
                sec_tab_comp.append(post_table_md)

            table_comps.append(
                SectionComponent(
                    title=tab_title,
                    contents=sec_tab_comp,
                )
            )

        # ignore the name as a parameter
        param_ids = [
            p.id for p in self._task.parent.parent["_parameters"].task if p.id != "name"
        ]
        if len(param_ids) > 0:
            # Extract parameter from the Parameter Task. That is less brittle.
            parameter_data = TaskToDict(
                only_repr=self._only_repr, runtime=self.runtime
            )(self._task.parent.parent["_parameters"].task, graph=self._graph)
            param_component = ArtifactsComponent(
                data=[parameter_data["data"][pid] for pid in param_ids]
            )
        else:
            param_component = TitleComponent(text="No Parameters")

        parameter_table = SectionComponent(
            title="Flow Parameters",
            contents=[param_component],
        ).render()

        # Don't include parameter ids + "name" in the task artifacts
        artifactlist = [
            task_data_dict["data"][k]
            for k in task_data_dict["data"]
            if k not in param_ids
        ]
        if len(artifactlist) > 0:
            artifact_component = ArtifactsComponent(data=artifactlist).render()
        else:
            artifact_component = TitleComponent(text="No Artifacts")

        artifact_section = SectionComponent(
            title="Artifacts", contents=[artifact_component]
        ).render()
        dag_component = SectionComponent(
            title="DAG", contents=[DagComponent(data=task_data_dict["graph"]).render()]
        ).render()

        page_contents = []
        if len(self._components) > 0:
            page_contents.extend(self._components)

        page_contents.extend(
            [
                metadata_table,
                parameter_table,
                artifact_section,
            ]
        )
        if len(table_comps) > 0:
            table_section = SectionComponent(
                title="Tabular Data", contents=table_comps
            ).render()
            page_contents.append(table_section)

        if len(img_components) > 0:
            img_section = SectionComponent(
                title="Image Data",
                columns=len(img_components),
                contents=img_components,
            ).render()
            page_contents.append(img_section)

        page_contents.append(dag_component)

        page_component = PageComponent(
            title=self._page_title,
            contents=page_contents,
        ).render()

        final_component_dict["components"].append(
            TitleComponent(text=task_data_dict["pathspec"]).render()
        )
        final_component_dict["components"].append(page_component)

        # These Properties will provide a way to access these components
        # once render is finished
        # this will Make this object reusable for run level cards.
        self.final_component = final_component_dict

        self.page_component = page_component

        return final_component_dict


class ErrorCard(MetaflowCard):

    type = "error"

    def __init__(self, options={}, components=[], graph=None):
        self._only_repr = True
        self._graph = None if graph is None else transform_flow_graph(graph)
        self._components = components

    def render(self, task, stack_trace=None):
        RENDER_TEMPLATE = read_file(RENDER_TEMPLATE_PATH)
        JS_DATA = read_file(JS_PATH)
        CSS_DATA = read_file(CSS_PATH)
        trace = "None"
        if stack_trace is not None:
            trace = stack_trace

        page = PageComponent(
            title="Error Card",
            contents=[
                SectionComponent(
                    title="Card Render Failed With Error",
                    contents=[LogComponent(data=trace)],
                )
            ],
        ).render()
        final_component_dict = dict(
            metadata={
                "pathspec": task.pathspec,
            },
            components=[page],
        )
        pt = self._get_mustache()
        data_dict = dict(
            task_data=base64.b64encode(
                json.dumps(final_component_dict).encode("utf-8")
            ).decode("utf-8"),
            javascript=JS_DATA,
            css=CSS_DATA,
            title=task.pathspec,
            card_data_id=uuid.uuid4(),
        )
        return pt.render(RENDER_TEMPLATE, data_dict)


class DefaultCardJSON(MetaflowCard):

    type = "default_json"

    def __init__(self, options=dict(only_repr=True), components=[], graph=None):
        self._only_repr = True
        self._graph = None if graph is None else transform_flow_graph(graph)
        if "only_repr" in options:
            self._only_repr = options["only_repr"]
        self._components = components

    def render(self, task):
        final_component_dict = TaskInfoComponent(
            task,
            only_repr=self._only_repr,
            graph=self._graph,
            components=self._components,
        ).render()
        return json.dumps(final_component_dict)


class DefaultCard(MetaflowCard):

    ALLOW_USER_COMPONENTS = True

    RUNTIME_UPDATABLE = True

    RELOAD_POLICY = MetaflowCard.RELOAD_POLICY_ONCHANGE

    type = "default"

    def __init__(self, options=dict(only_repr=True), components=[], graph=None):
        self._only_repr = True
        self._graph = None if graph is None else transform_flow_graph(graph)
        if "only_repr" in options:
            self._only_repr = options["only_repr"]
        self._components = components

    def render(self, task, runtime=False):
        RENDER_TEMPLATE = read_file(RENDER_TEMPLATE_PATH)
        JS_DATA = read_file(JS_PATH)
        CSS_DATA = read_file(CSS_PATH)
        final_component_dict = TaskInfoComponent(
            task,
            only_repr=self._only_repr,
            graph=self._graph,
            components=self._components,
            runtime=runtime,
        ).render()
        pt = self._get_mustache()
        data_dict = dict(
            task_data=base64.b64encode(
                json.dumps(final_component_dict).encode("utf-8")
            ).decode("utf-8"),
            javascript=JS_DATA,
            title=task.pathspec,
            css=CSS_DATA,
            card_data_id=uuid.uuid4(),
            RENDER_COMPLETE=not runtime,
        )
        return pt.render(RENDER_TEMPLATE, data_dict)

    def render_runtime(self, task, data):
        return self.render(task, runtime=True)

    def refresh(self, task, data):
        return data["components"]

    def reload_content_token(self, task, data):
        """
        The reload token will change when the component array has changed in the Metaflow card.
        The change in the component array is signified by the change in the component_update_ts.
        """
        if task.finished:
            return "final"
        # `component_update_ts` will never be None. It is set to a default value when the `ComponentStore` is instantiated
        # And it is updated when components added / removed / changed from the `ComponentStore`.
        return "runtime-%s" % (str(data["component_update_ts"]))


class BlankCard(MetaflowCard):

    ALLOW_USER_COMPONENTS = True

    RUNTIME_UPDATABLE = True

    RELOAD_POLICY = MetaflowCard.RELOAD_POLICY_ONCHANGE

    type = "blank"

    def __init__(self, options=dict(title=""), components=[], graph=None):
        self._graph = None if graph is None else transform_flow_graph(graph)
        self._title = ""
        if "title" in options:
            self._title = options["title"]
        self._components = components

    def render(self, task, components=[], runtime=False):
        RENDER_TEMPLATE = read_file(RENDER_TEMPLATE_PATH)
        JS_DATA = read_file(JS_PATH)
        CSS_DATA = read_file(CSS_PATH)
        if type(components) != list:
            components = []
        page_component = PageComponent(
            title=self._title,
            contents=components + self._components,
        ).render()
        final_component_dict = dict(
            metadata={
                "pathspec": task.pathspec,
            },
            components=[page_component],
        )
        pt = self._get_mustache()
        data_dict = dict(
            task_data=base64.b64encode(
                json.dumps(final_component_dict).encode("utf-8")
            ).decode("utf-8"),
            javascript=JS_DATA,
            title=task.pathspec,
            css=CSS_DATA,
            card_data_id=uuid.uuid4(),
            RENDER_COMPLETE=not runtime,
        )
        return pt.render(RENDER_TEMPLATE, data_dict)

    def render_runtime(self, task, data):
        return self.render(task, runtime=True)

    def refresh(self, task, data):
        return data["components"]

    def reload_content_token(self, task, data):
        """
        The reload token will change when the component array has changed in the Metaflow card.
        The change in the component array is signified by the change in the component_update_ts.
        """
        if task.finished:
            return "final"
        # `component_update_ts` will never be None. It is set to a default value when the `ComponentStore` is instantiated
        # And it is updated when components added / removed / changed from the `ComponentStore`.
        return "runtime-%s" % (str(data["component_update_ts"]))


class TaskSpecCard(MetaflowCard):
    type = "taskspec_card"

    def render(self, task):
        return "%s" % task.pathspec



import atexit
import importlib
import itertools
import pickle
import re
import sys

from .consts import OP_CALLFUNC, OP_GETVAL, OP_SETVAL
from .client import Client
from .override_decorators import LocalException
from .utils import get_canonical_name


def _clean_client(client):
    client.cleanup()


class _WrappedModule(object):
    def __init__(self, loader, prefix, exports, exception_classes, client):
        self._loader = loader
        self._prefix = prefix
        self._client = client
        is_match = re.compile(
            r"^%s\.([a-zA-Z_][a-zA-Z0-9_]*)$" % prefix.replace(".", r"\.")  # noqa W605
        )
        self._exports = {}
        self._aliases = exports["aliases"]
        for k in ("classes", "functions", "values"):
            result = []
            for item in exports[k]:
                m = is_match.match(item)
                if m:
                    result.append(m.group(1))
            self._exports[k] = result
        self._exception_classes = {}
        for k, v in exception_classes.items():
            m = is_match.match(k)
            if m:
                self._exception_classes[m.group(1)] = v

    def __getattr__(self, name):
        if name == "__loader__":
            return self._loader
        if name in ("__name__", "__package__"):
            return self._prefix
        if name in ("__file__", "__path__"):
            return self._client.name

        # Make the name canonical because the prefix is also canonical.
        name = get_canonical_name(self._prefix + "." + name, self._aliases)[
            len(self._prefix) + 1 :
        ]
        if name in self._exports["classes"]:
            # We load classes lazily
            return self._client.get_local_class("%s.%s" % (self._prefix, name))
        elif name in self._exports["functions"]:
            # TODO: Grab doc back from the remote side like in _make_method
            def func(*args, **kwargs):
                return self._client.stub_request(
                    None, OP_CALLFUNC, "%s.%s" % (self._prefix, name), *args, **kwargs
                )

            func.__name__ = name
            func.__doc__ = "Unknown (TODO)"
            return func
        elif name in self._exports["values"]:
            return self._client.stub_request(
                None, OP_GETVAL, "%s.%s" % (self._prefix, name)
            )
        elif name in self._exception_classes:
            return self._exception_classes[name]
        else:
            # Try to see if this is a submodule that we can load
            m = None
            try:
                m = self._loader.load_module(".".join([self._prefix, name]))
            except ImportError:
                pass
            if m is None:
                raise AttributeError(
                    "module '%s' has no attribute '%s' -- contact the author of the "
                    "configuration if this is something "
                    "you expect to work (support may be added if it exists in the "
                    "original library)" % (self._prefix, name)
                )
            return m

    def __setattr__(self, name, value):
        if name in (
            "package",
            "__spec__",
            "_loader",
            "_prefix",
            "_client",
            "_exports",
            "_exception_classes",
            "_aliases",
        ):
            object.__setattr__(self, name, value)
            return
        if isinstance(value, _WrappedModule):
            # This is when a module sets itself as an attribute of another
            # module when loading
            object.__setattr__(self, name, value)
            return

        # Make the name canonical because the prefix is also canonical.
        name = get_canonical_name(self._prefix + "." + name, self._aliases)[
            len(self._prefix) + 1 :
        ]
        if name in self._exports["values"]:
            self._client.stub_request(
                None, OP_SETVAL, "%s.%s" % (self._prefix, name), value
            )
        elif name in self._exports["classes"] or name in self._exports["functions"]:
            raise ValueError
        else:
            raise AttributeError(name)


class ModuleImporter(object):
    # This ModuleImporter implements the Importer Protocol defined in PEP 302
    def __init__(
        self,
        python_executable,
        pythonpath,
        max_pickle_version,
        config_dir,
        module_prefixes,
    ):
        self._module_prefixes = module_prefixes
        self._python_executable = python_executable
        self._pythonpath = pythonpath
        self._config_dir = config_dir
        self._client = None
        self._max_pickle_version = max_pickle_version
        self._handled_modules = None
        self._aliases = {}

    def find_module(self, fullname, path=None):
        if self._handled_modules is not None:
            if get_canonical_name(fullname, self._aliases) in self._handled_modules:
                return self
            return None
        if any([fullname.startswith(prefix) for prefix in self._module_prefixes]):
            # We potentially handle this
            return self
        return None

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        if self._client is None:
            if sys.version_info[0] < 3:
                raise NotImplementedError(
                    "Environment escape imports are not supported in Python 2"
                )
            # We initialize a client and query the modules we handle
            # The max_pickle_version is the pickle version that the server (so
            # the underlying interpreter we call into) supports; we determine
            # what version the current environment support and take the minimum
            # of those two
            max_pickle_version = min(self._max_pickle_version, pickle.HIGHEST_PROTOCOL)

            self._client = Client(
                self._module_prefixes,
                self._python_executable,
                self._pythonpath,
                max_pickle_version,
                self._config_dir,
            )
            atexit.register(_clean_client, self._client)

            # Get information about overrides and what the server knows about
            exports = self._client.get_exports()
            ex_overrides = self._client.get_local_exception_overrides()

            prefixes = set()
            export_classes = exports.get("classes", [])
            export_functions = exports.get("functions", [])
            export_values = exports.get("values", [])
            export_exceptions = exports.get("exceptions", [])
            self._aliases = exports.get("aliases", {})
            for name in itertools.chain(
                export_classes, export_functions, export_values
            ):
                splits = name.rsplit(".", 1)
                prefixes.add(splits[0])

            # Now look at the exceptions coming from the server
            formed_exception_classes = {}
            for ex_name, ex_parents in export_exceptions:
                # Exception is a tuple (name, (parents,))
                # Exceptions are also given in order of instantiation (ie: the
                # server already topologically sorted them)
                ex_class_dict = ex_overrides.get(ex_name, None)
                if ex_class_dict is None:
                    ex_class_dict = {}
                else:
                    ex_class_dict = dict(ex_class_dict.__dict__)
                parents = []
                for fake_base in ex_parents:
                    if fake_base.startswith("builtins."):
                        # This is something we know of here
                        parents.append(eval(fake_base[9:]))
                    else:
                        # It's in formed_classes
                        parents.append(formed_exception_classes[fake_base])
                splits = ex_name.rsplit(".", 1)
                ex_class_dict["__user_defined__"] = set(ex_class_dict.keys())
                new_class = type(splits[1], tuple(parents), ex_class_dict)
                new_class.__module__ = splits[0]
                new_class.__name__ = splits[1]
                formed_exception_classes[ex_name] = new_class

            # Now update prefixes as needed
            for name in formed_exception_classes:
                splits = name.rsplit(".", 1)
                prefixes.add(splits[0])

            # We will make sure that we create modules even for "empty" prefixes
            # because packages are always loaded hierarchically so if we have
            # something in `a.b.c` but nothing directly in `a`, we still need to
            # create a module named `a`. There is probably a better way of doing this
            all_prefixes = list(prefixes)
            for prefix in all_prefixes:
                parts = prefix.split(".")
                cur = parts[0]
                for i in range(1, len(parts)):
                    prefixes.add(cur)
                    cur = ".".join([cur, parts[i]])

            # We now know all the modules that we can handle. We update
            # handled_module and return the module if we have it or raise ImportError
            self._handled_modules = {}
            for prefix in prefixes:
                self._handled_modules[prefix] = _WrappedModule(
                    self, prefix, exports, formed_exception_classes, self._client
                )
        canonical_fullname = get_canonical_name(fullname, self._aliases)
        # Modules are created canonically but we need to return something for any
        # of the aliases.
        module = self._handled_modules.get(canonical_fullname)
        if module is None:
            raise ImportError
        sys.modules[fullname] = module
        return module


def create_modules(python_executable, pythonpath, max_pickle_version, path, prefixes):
    # This is an extra verification to make sure we are not trying to use the
    # environment escape for something that is in the system
    for prefix in prefixes:
        try:
            importlib.import_module(prefix)
        except ImportError:
            pass
        else:
            # pass
            raise RuntimeError(
                "Trying to override %s when module exists in system" % prefix
            )

    # The first version forces the use of the environment escape even if the module
    # exists in the system. This is useful for testing to make sure that the
    # environment escape is used. The second version is more production friendly and
    # will only use the environment escape if the module cannot be found

    # sys.meta_path.insert(0, ModuleImporter(python_path, path, prefixes))
    sys.meta_path.append(
        ModuleImporter(
            python_executable, pythonpath, max_pickle_version, path, prefixes
        )
    )



""" 

"""

from collections import namedtuple
from io import BytesIO
import os
import json
import shutil

from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.metaflow_config import (
    CARD_S3ROOT,
    CARD_LOCALROOT,
    DATASTORE_LOCAL_DIR,
    CARD_SUFFIX,
    CARD_AZUREROOT,
    CARD_GSROOT,
    SKIP_CARD_DUALWRITE,
)
import metaflow.metaflow_config as metaflow_config

from .exception import CardNotPresentException

TEMP_DIR_NAME = "metaflow_card_cache"
NUM_SHORT_HASH_CHARS = 5

CardInfo = namedtuple("CardInfo", ["type", "hash", "id", "filename"])


class CardNameSuffix:
    DATA = "data.json"
    CARD = "html"


class CardPathSuffix:
    DATA = "runtime"
    CARD = "cards"


def path_spec_resolver(pathspec):
    splits = pathspec.split("/")
    splits.extend([None] * (4 - len(splits)))
    return tuple(splits)


def is_file_present(path):
    try:
        os.stat(path)
        return True
    except FileNotFoundError:
        return False
    except:
        raise


class CardDatastore(object):
    @classmethod
    def get_storage_root(cls, storage_type):
        if storage_type == "s3":
            return CARD_S3ROOT
        elif storage_type == "azure":
            return CARD_AZUREROOT
        elif storage_type == "gs":
            return CARD_GSROOT
        elif storage_type == "local":
            # Borrowing some of the logic from LocalStorage.get_storage_root
            result = CARD_LOCALROOT
            if result is None:
                current_path = os.getcwd()
                check_dir = os.path.join(current_path, DATASTORE_LOCAL_DIR, CARD_SUFFIX)
                check_dir = os.path.realpath(check_dir)
                orig_path = check_dir
                while not os.path.isdir(check_dir):
                    new_path = os.path.dirname(current_path)
                    if new_path == current_path:
                        break  # We are no longer making upward progress
                    current_path = new_path
                    check_dir = os.path.join(
                        current_path, DATASTORE_LOCAL_DIR, CARD_SUFFIX
                    )
                result = orig_path

            return result
        else:
            # Let's make it obvious we need to update this block for each new datastore backend...
            raise NotImplementedError(
                "Card datastore does not support backend %s" % (storage_type,)
            )

    def __init__(self, flow_datastore, pathspec=None):
        self._backend = flow_datastore._storage_impl
        self._flow_name = flow_datastore.flow_name
        _, run_id, step_name, _ = pathspec.split("/")
        self._run_id = run_id
        self._step_name = step_name
        self._pathspec = pathspec
        self._temp_card_save_path = self._get_card_write_path(base_pth=TEMP_DIR_NAME)

    @classmethod
    def get_card_location(
        cls, base_path, card_name, uuid, card_id=None, suffix=CardNameSuffix.CARD
    ):
        chash = uuid
        if card_id is None:
            card_file_name = "%s-%s.%s" % (card_name, chash, suffix)
        else:
            card_file_name = "%s-%s-%s.%s" % (card_name, card_id, chash, suffix)
        return os.path.join(base_path, card_file_name)

    def _make_path(
        self, base_pth, pathspec=None, with_steps=False, suffix=CardPathSuffix.CARD
    ):
        sysroot = base_pth
        if pathspec is not None:
            # since most cards are at a task level there will always be 4 non-none values returned
            flow_name, run_id, step_name, task_id = path_spec_resolver(pathspec)

        # We have a condition that checks for `with_steps` because
        # when cards were introduced there was an assumption made
        # about task-ids being unique.
        # This assumption is incorrect since pathspec needs to be
        # unique but there is no such guarantees on task-ids.
        # This is why we have a `with_steps` flag that allows
        # constructing the path with and without steps so that
        # older-cards (cards with a path without `steps/<stepname>` in them)
        # can also be accessed by the card cli and the card client.
        if with_steps:
            pth_arr = [
                sysroot,
                flow_name,
                "runs",
                run_id,
                "steps",
                step_name,
                "tasks",
                task_id,
                suffix,
            ]
        else:
            pth_arr = [
                sysroot,
                flow_name,
                "runs",
                run_id,
                "tasks",
                task_id,
                suffix,
            ]
        if sysroot == "" or sysroot is None:
            pth_arr.pop(0)
        return os.path.join(*pth_arr)

    def _get_data_read_path(self, base_pth=""):
        return self._make_path(
            base_pth=base_pth,
            pathspec=self._pathspec,
            with_steps=True,
            suffix=CardPathSuffix.DATA,
        )

    def _get_data_write_path(self, base_pth=""):
        return self._make_path(
            base_pth=base_pth,
            pathspec=self._pathspec,
            with_steps=True,
            suffix=CardPathSuffix.DATA,
        )

    def _get_card_write_path(
        self,
        base_pth="",
    ):
        return self._make_path(
            base_pth,
            pathspec=self._pathspec,
            with_steps=True,
            suffix=CardPathSuffix.CARD,
        )

    def _get_card_read_path(self, base_pth="", with_steps=False):
        return self._make_path(
            base_pth,
            pathspec=self._pathspec,
            with_steps=with_steps,
            suffix=CardPathSuffix.CARD,
        )

    @staticmethod
    def info_from_path(path, suffix=CardNameSuffix.CARD):
        """
        Args:
            path (str): The path to the card

        Raises:
            Exception: When the card_path is invalid

        Returns:
            CardInfo
        """
        card_file_name = path.split("/")[-1]
        file_split = card_file_name.split("-")

        if len(file_split) not in [2, 3]:
            raise Exception(
                "Invalid file name %s. Card/Data file names should be of form TYPE-HASH.%s or TYPE-ID-HASH.%s"
                % (card_file_name, suffix, suffix)
            )
        card_type, card_hash, card_id = None, None, None

        if len(file_split) == 2:
            card_type, card_hash = file_split
        else:
            card_type, card_id, card_hash = file_split

        card_hash = card_hash.split("." + suffix)[0]
        return CardInfo(card_type, card_hash, card_id, card_file_name)

    def save_data(self, uuid, card_type, json_data, card_id=None):
        card_file_name = card_type
        loc = self.get_card_location(
            self._get_data_write_path(),
            card_file_name,
            uuid,
            card_id=card_id,
            suffix=CardNameSuffix.DATA,
        )
        self._backend.save_bytes(
            [(loc, BytesIO(json.dumps(json_data).encode("utf-8")))], overwrite=True
        )

    def save_card(self, uuid, card_type, card_html, card_id=None, overwrite=True):
        card_file_name = card_type
        # TEMPORARY_WORKAROUND: FIXME (LATER) : Fix the duplication of below block in a few months.
        # Check file blame to understand the age of this temporary workaround.

        # This function will end up saving cards at two locations.
        # Thereby doubling the number of cards. (Which is a temporary fix)
        # Why do this ? :
        # When cards were introduced there was an assumption made about task-ids being unique.
        # This assumption was incorrect.
        # Only the pathspec needs to be unique but there is no such guarantees about task-ids.
        # When task-ids are non-unique, card read would result in finding incorrect cards.
        # This happens because cards were stored based on task-ids.
        # If we immediately switch from storing based on task-ids to a step-name abstraction folder,
        # then card reading will crash for many users.
        # It would especially happen for users who are accessing cards created by a newer
        # MF client from an older version of MF client.
        # It will also easily end up breaking the metaflow-ui (which maybe using a client from an older version).
        # Hence, we are writing cards to both paths so that we can introduce breaking changes later in the future.
        card_path_with_steps = self.get_card_location(
            self._get_card_write_path(),
            card_file_name,
            uuid,
            card_id=card_id,
            suffix=CardNameSuffix.CARD,
        )
        if SKIP_CARD_DUALWRITE:
            self._backend.save_bytes(
                [(card_path_with_steps, BytesIO(bytes(card_html, "utf-8")))],
                overwrite=overwrite,
            )
        else:
            card_path_without_steps = self.get_card_location(
                self._get_card_read_path(with_steps=False),
                card_file_name,
                uuid,
                card_id=card_id,
                suffix=CardNameSuffix.CARD,
            )
            for cp in [card_path_with_steps, card_path_without_steps]:
                self._backend.save_bytes(
                    [(cp, BytesIO(bytes(card_html, "utf-8")))], overwrite=overwrite
                )

        return self.info_from_path(card_path_with_steps, suffix=CardNameSuffix.CARD)

    def _list_card_paths(self, card_type=None, card_hash=None, card_id=None):
        # Check for new cards first
        card_paths = []
        card_paths_with_steps = self._backend.list_content(
            [self._get_card_read_path(with_steps=True)]
        )

        if len(card_paths_with_steps) == 0:
            card_paths_without_steps = self._backend.list_content(
                [self._get_card_read_path(with_steps=False)]
            )
            if len(card_paths_without_steps) == 0:
                # If there are no files found on the Path then raise an error of
                raise CardNotPresentException(
                    self._pathspec,
                    card_hash=card_hash,
                    card_type=card_type,
                )
            else:
                card_paths = card_paths_without_steps
        else:
            card_paths = card_paths_with_steps

        cards_found = []
        for task_card_path in card_paths:
            card_path = task_card_path.path
            card_info = self.info_from_path(card_path, suffix=CardNameSuffix.CARD)
            if card_type is not None and card_info.type != card_type:
                continue
            elif card_hash is not None:
                if not card_info.hash.startswith(card_hash):
                    continue
            elif card_id is not None and card_info.id != card_id:
                continue

            if task_card_path.is_file:
                cards_found.append(card_path)

        return cards_found

    def _list_card_data(self, card_type=None, card_hash=None, card_id=None):
        card_data_paths = self._backend.list_content([self._get_data_read_path()])
        data_found = []

        for data_path in card_data_paths:
            _pth = data_path.path
            card_info = self.info_from_path(_pth, suffix=CardNameSuffix.DATA)
            if card_type is not None and card_info.type != card_type:
                continue
            elif card_hash is not None:
                if not card_info.hash.startswith(card_hash):
                    continue
            elif card_id is not None and card_info.id != card_id:
                continue
            if data_path.is_file:
                data_found.append(_pth)

        return data_found

    def create_full_path(self, card_path):
        return os.path.join(self._backend.datastore_root, card_path)

    def get_card_names(self, card_paths):
        return [
            self.info_from_path(path, suffix=CardNameSuffix.CARD) for path in card_paths
        ]

    def get_card_html(self, path):
        with self._backend.load_bytes([path]) as get_results:
            for _, path, _ in get_results:
                if path is not None:
                    with open(path, "r") as f:
                        return f.read()

    def get_card_data(self, path):
        with self._backend.load_bytes([path]) as get_results:
            for _, path, _ in get_results:
                if path is not None:
                    with open(path, "r") as f:
                        return json.loads(f.read())

    def cache_locally(self, path, save_path=None):
        """
        Saves the data present in the `path` the `metaflow_card_cache` directory or to the `save_path`.
        """
        # todo : replace this function with the FileCache
        if save_path is None:
            if not is_file_present(self._temp_card_save_path):
                LocalStorage._makedirs(self._temp_card_save_path)
        else:
            save_dir = os.path.dirname(save_path)
            if save_dir != "" and not is_file_present(save_dir):
                LocalStorage._makedirs(os.path.dirname(save_path))

        with self._backend.load_bytes([path]) as get_results:
            for key, path, meta in get_results:
                if path is not None:
                    main_path = path
                    if save_path is None:
                        file_name = key.split("/")[-1]
                        main_path = os.path.join(self._temp_card_save_path, file_name)
                    else:
                        main_path = save_path
                    shutil.copy(path, main_path)
                    return main_path

    def extract_data_paths(self, card_type=None, card_hash=None, card_id=None):
        return self._list_card_data(
            # card_hash is the unique identifier to the card.
            # Its no longer the actual hash!
            card_type=card_type,
            card_hash=card_hash,
            card_id=card_id,
        )

    def extract_card_paths(self, card_type=None, card_hash=None, card_id=None):
        return self._list_card_paths(
            card_type=card_type, card_hash=card_hash, card_id=card_id
        )



from metaflow.decorators import StepDecorator
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import MAX_ATTEMPTS


class RetryDecorator(StepDecorator):
    """
    Specifies the number of times the task corresponding
    to a step needs to be retried.

    This decorator is useful for handling transient errors, such as networking issues.
    If your task contains operations that can't be retried safely, e.g. database updates,
    it is advisable to annotate it with `@retry(times=0)`.

    This can be used in conjunction with the `@catch` decorator. The `@catch`
    decorator will execute a no-op task after all retries have been exhausted,
    ensuring that the flow execution can continue.

    Parameters
    ----------
    times : int, default: 3
        Number of times to retry this task.
    minutes_between_retries : int, default: 2
        Number of minutes between retries.
    """

    name = "retry"
    defaults = {"times": "3", "minutes_between_retries": "2"}

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # The total number of attempts must not exceed MAX_ATTEMPTS.
        # attempts = normal task (1) + retries (N) + @catch fallback (1)
        if int(self.attributes["times"]) + 2 > MAX_ATTEMPTS:
            raise MetaflowException(
                "The maximum number of retries is "
                "@retry(times=%d)." % (MAX_ATTEMPTS - 2)
            )

    def step_task_retry_count(self):
        return int(self.attributes["times"]), 0



import json
import os
import random
import string
import sys
from datetime import datetime, timedelta
from io import BytesIO

import metaflow.util as util
from metaflow import current
from metaflow.decorators import flow_decorators
from metaflow.exception import MetaflowException
from metaflow.includefile import FilePathClass
from metaflow.metaflow_config import (
    AIRFLOW_KUBERNETES_CONN_ID,
    AIRFLOW_KUBERNETES_KUBECONFIG_CONTEXT,
    AIRFLOW_KUBERNETES_KUBECONFIG_FILE,
    AIRFLOW_KUBERNETES_STARTUP_TIMEOUT_SECONDS,
    AWS_SECRETS_MANAGER_DEFAULT_REGION,
    AZURE_STORAGE_BLOB_SERVICE_ENDPOINT,
    CARD_AZUREROOT,
    CARD_GSROOT,
    CARD_S3ROOT,
    DATASTORE_SYSROOT_AZURE,
    DATASTORE_SYSROOT_GS,
    DATASTORE_SYSROOT_S3,
    DATATOOLS_S3ROOT,
    DEFAULT_SECRETS_BACKEND_TYPE,
    KUBERNETES_SECRETS,
    KUBERNETES_SERVICE_ACCOUNT,
    S3_ENDPOINT_URL,
    SERVICE_HEADERS,
    SERVICE_INTERNAL_URL,
)

from metaflow.metaflow_config_funcs import config_values

from metaflow.parameters import (
    DelayedEvaluationParameter,
    JSONTypeClass,
    deploy_time_eval,
)

# TODO: Move chevron to _vendor
from metaflow.plugins.cards.card_modules import chevron
from metaflow.plugins.kubernetes.kubernetes import Kubernetes
from metaflow.plugins.timeout_decorator import get_run_time_limit_for_task
from metaflow.util import compress_list, dict_to_cli_options, get_username

from . import airflow_utils
from .airflow_utils import AIRFLOW_MACROS, TASK_ID_XCOM_KEY, AirflowTask, Workflow
from .exception import AirflowException
from .sensors import SUPPORTED_SENSORS

AIRFLOW_DEPLOY_TEMPLATE_FILE = os.path.join(os.path.dirname(__file__), "dag.py")


class Airflow(object):
    TOKEN_STORAGE_ROOT = "mf.airflow"

    def __init__(
        self,
        name,
        graph,
        flow,
        code_package_sha,
        code_package_url,
        metadata,
        flow_datastore,
        environment,
        event_logger,
        monitor,
        production_token,
        tags=None,
        namespace=None,
        username=None,
        max_workers=None,
        worker_pool=None,
        description=None,
        file_path=None,
        workflow_timeout=None,
        is_paused_upon_creation=True,
    ):
        self.name = name
        self.graph = graph
        self.flow = flow
        self.code_package_sha = code_package_sha
        self.code_package_url = code_package_url
        self.metadata = metadata
        self.flow_datastore = flow_datastore
        self.environment = environment
        self.event_logger = event_logger
        self.monitor = monitor
        self.tags = tags
        self.namespace = namespace  # this is the username space
        self.username = username
        self.max_workers = max_workers
        self.description = description
        self._depends_on_upstream_sensors = False
        self._file_path = file_path
        _, self.graph_structure = self.graph.output_steps()
        self.worker_pool = worker_pool
        self.is_paused_upon_creation = is_paused_upon_creation
        self.workflow_timeout = workflow_timeout
        self.schedule = self._get_schedule()
        self.parameters = self._process_parameters()
        self.production_token = production_token
        self.contains_foreach = self._contains_foreach()

    @classmethod
    def get_existing_deployment(cls, name, flow_datastore):
        _backend = flow_datastore._storage_impl
        token_exists = _backend.is_file([cls.get_token_path(name)])
        if not token_exists[0]:
            return None
        with _backend.load_bytes([cls.get_token_path(name)]) as get_results:
            for _, path, _ in get_results:
                if path is not None:
                    with open(path, "r") as f:
                        data = json.loads(f.read())
                    return (data["owner"], data["production_token"])

    @classmethod
    def get_token_path(cls, name):
        return os.path.join(cls.TOKEN_STORAGE_ROOT, name)

    @classmethod
    def save_deployment_token(cls, owner, name, token, flow_datastore):
        _backend = flow_datastore._storage_impl
        _backend.save_bytes(
            [
                (
                    cls.get_token_path(name),
                    BytesIO(
                        bytes(
                            json.dumps({"production_token": token, "owner": owner}),
                            "utf-8",
                        )
                    ),
                )
            ],
            overwrite=False,
        )

    def _get_schedule(self):
        # Using the cron presets provided here :
        # https://airflow.apache.org/docs/apache-airflow/stable/dag-run.html?highlight=schedule%20interval#cron-presets
        schedule = self.flow._flow_decorators.get("schedule")
        if not schedule:
            return None
        schedule = schedule[0]
        if schedule.attributes["cron"]:
            return schedule.attributes["cron"]
        elif schedule.attributes["weekly"]:
            return "@weekly"
        elif schedule.attributes["hourly"]:
            return "@hourly"
        elif schedule.attributes["daily"]:
            return "@daily"
        return None

    def _get_retries(self, node):
        max_user_code_retries = 0
        max_error_retries = 0
        foreach_default_retry = 1
        # Different decorators may have different retrying strategies, so take
        # the max of them.
        for deco in node.decorators:
            user_code_retries, error_retries = deco.step_task_retry_count()
            max_user_code_retries = max(max_user_code_retries, user_code_retries)
            max_error_retries = max(max_error_retries, error_retries)
        parent_is_foreach = any(  # The immediate parent is a foreach node.
            self.graph[n].type == "foreach" for n in node.in_funcs
        )

        if parent_is_foreach:
            max_user_code_retries + foreach_default_retry
        return max_user_code_retries, max_user_code_retries + max_error_retries

    def _get_retry_delay(self, node):
        retry_decos = [deco for deco in node.decorators if deco.name == "retry"]
        if len(retry_decos) > 0:
            retry_mins = retry_decos[0].attributes["minutes_between_retries"]
            return timedelta(minutes=int(retry_mins))
        return None

    def _process_parameters(self):
        airflow_params = []
        type_transform_dict = {
            int.__name__: "integer",
            str.__name__: "string",
            bool.__name__: "string",
            float.__name__: "number",
        }

        for var, param in self.flow._get_parameters():
            # Airflow requires defaults set for parameters.
            value = deploy_time_eval(param.kwargs.get("default"))
            # Setting airflow related param args.
            airflow_param = dict(
                name=param.name,
            )
            if value is not None:
                airflow_param["default"] = value
            if param.kwargs.get("help"):
                airflow_param["description"] = param.kwargs.get("help")

            # Since we will always have a default value and `deploy_time_eval` resolved that to an actual value
            # we can just use the `default` to infer the object's type.
            # This avoids parsing/identifying types like `JSONType` or `FilePathClass`
            # which are returned by calling `param.kwargs.get("type")`
            param_type = type(airflow_param["default"])

            # extract the name of the type and resolve the type-name
            # compatible with Airflow.
            param_type_name = getattr(param_type, "__name__", None)
            if param_type_name in type_transform_dict:
                airflow_param["type"] = type_transform_dict[param_type_name]

            if param_type_name == bool.__name__:
                airflow_param["default"] = str(airflow_param["default"])

            airflow_params.append(airflow_param)

        return airflow_params

    def _compress_input_path(
        self,
        steps,
    ):
        """
        This function is meant to compress the input paths, and it specifically doesn't use
        `metaflow.util.compress_list` under the hood. The reason is that the `AIRFLOW_MACROS.RUN_ID` is a complicated
        macro string that doesn't behave nicely with `metaflow.util.decompress_list`, since the `decompress_util`
        function expects a string which doesn't contain any delimiter characters and the run-id string does. Hence, we
        have a custom compression string created via `_compress_input_path` function instead of `compress_list`.
        """
        return "%s:" % (AIRFLOW_MACROS.RUN_ID) + ",".join(
            self._make_input_path(step, only_task_id=True) for step in steps
        )

    def _make_foreach_input_path(self, step_name):
        return (
            "%s/%s/:{{ task_instance.xcom_pull(task_ids='%s',key='%s') | join_list }}"
            % (
                AIRFLOW_MACROS.RUN_ID,
                step_name,
                step_name,
                TASK_ID_XCOM_KEY,
            )
        )

    def _make_input_path(self, step_name, only_task_id=False):
        """
        This is set using the `airflow_internal` decorator to help pass state.
        This will pull the `TASK_ID_XCOM_KEY` xcom which holds task-ids.
        The key is set via the `MetaflowKubernetesOperator`.
        """
        task_id_string = "/%s/{{ task_instance.xcom_pull(task_ids='%s',key='%s') }}" % (
            step_name,
            step_name,
            TASK_ID_XCOM_KEY,
        )

        if only_task_id:
            return task_id_string

        return "%s%s" % (AIRFLOW_MACROS.RUN_ID, task_id_string)

    def _to_job(self, node):
        """
        This function will transform the node's specification into Airflow compatible operator arguments.
        Since this function is long, below is the summary of the two major duties it performs:
            1. Based on the type of the graph node (start/linear/foreach/join etc.)
                it will decide how to set the input paths
            2. Based on node's decorator specification convert the information into
                a job spec for the KubernetesPodOperator.
        """
        # Add env vars from the optional @environment decorator.
        env_deco = [deco for deco in node.decorators if deco.name == "environment"]
        env = {}
        if env_deco:
            env = env_deco[0].attributes["vars"].copy()

        # The below if/else block handles "input paths".
        # Input Paths help manage dataflow across the graph.
        if node.name == "start":
            # POSSIBLE_FUTURE_IMPROVEMENT:
            # We can extract metadata about the possible upstream sensor triggers.
            # There is a previous commit (7bdf6) in the `airflow` branch that has `SensorMetaExtractor` class and
            # associated MACRO we have built to handle this case if a metadata regarding the sensor is needed.
            # Initialize parameters for the flow in the `start` step.
            # `start` step has no upstream input dependencies aside from
            # parameters.

            if len(self.parameters):
                env["METAFLOW_PARAMETERS"] = AIRFLOW_MACROS.PARAMETERS
            input_paths = None
        else:
            # If it is not the start node then we check if there are many paths
            # converging into it or a single path. Based on that we set the INPUT_PATHS
            if node.parallel_foreach:
                raise AirflowException(
                    "Parallel steps are not supported yet with Airflow."
                )
            is_foreach_join = (
                node.type == "join"
                and self.graph[node.split_parents[-1]].type == "foreach"
            )
            if is_foreach_join:
                input_paths = self._make_foreach_input_path(node.in_funcs[0])

            elif len(node.in_funcs) == 1:
                # set input paths where this is only one parent node
                # The parent-task-id is passed via the xcom; There is no other way to get that.
                # One key thing about xcoms is that they are immutable and only accepted if the task
                # doesn't fail.
                # From airflow docs :
                # "Note: If the first task run is not succeeded then on every retry task
                # XComs will be cleared to make the task run idempotent."
                input_paths = self._make_input_path(node.in_funcs[0])
            else:
                # this is a split scenario where there can be more than one input paths.
                input_paths = self._compress_input_path(node.in_funcs)

            # env["METAFLOW_INPUT_PATHS"] = input_paths

        env["METAFLOW_CODE_URL"] = self.code_package_url
        env["METAFLOW_FLOW_NAME"] = self.flow.name
        env["METAFLOW_STEP_NAME"] = node.name
        env["METAFLOW_OWNER"] = self.username

        metadata_env = self.metadata.get_runtime_environment("airflow")
        env.update(metadata_env)

        metaflow_version = self.environment.get_environment_info()
        metaflow_version["flow_name"] = self.graph.name
        metaflow_version["production_token"] = self.production_token
        env["METAFLOW_VERSION"] = json.dumps(metaflow_version)

        # Temporary passing of *some* environment variables. Do not rely on this
        # mechanism as it will be removed in the near future
        env.update(
            {
                k: v
                for k, v in config_values()
                if k.startswith("METAFLOW_CONDA_") or k.startswith("METAFLOW_DEBUG_")
            }
        )

        # Extract the k8s decorators for constructing the arguments of the K8s Pod Operator on Airflow.
        k8s_deco = [deco for deco in node.decorators if deco.name == "kubernetes"][0]
        user_code_retries, _ = self._get_retries(node)
        retry_delay = self._get_retry_delay(node)
        # This sets timeouts for @timeout decorators.
        # The timeout is set as "execution_timeout" for an airflow task.
        runtime_limit = get_run_time_limit_for_task(node.decorators)

        k8s = Kubernetes(self.flow_datastore, self.metadata, self.environment)
        user = util.get_username()

        labels = {
            "app": "metaflow",
            "app.kubernetes.io/name": "metaflow-task",
            "app.kubernetes.io/part-of": "metaflow",
            "app.kubernetes.io/created-by": user,
            # Question to (savin) : Should we have username set over here for created by since it is the
            # airflow installation that is creating the jobs.
            # Technically the "user" is the stakeholder but should these labels be present.
        }
        additional_mf_variables = {
            "METAFLOW_CODE_SHA": self.code_package_sha,
            "METAFLOW_CODE_URL": self.code_package_url,
            "METAFLOW_CODE_DS": self.flow_datastore.TYPE,
            "METAFLOW_USER": user,
            "METAFLOW_SERVICE_URL": SERVICE_INTERNAL_URL,
            "METAFLOW_SERVICE_HEADERS": json.dumps(SERVICE_HEADERS),
            "METAFLOW_DATASTORE_SYSROOT_S3": DATASTORE_SYSROOT_S3,
            "METAFLOW_DATATOOLS_S3ROOT": DATATOOLS_S3ROOT,
            "METAFLOW_DEFAULT_DATASTORE": self.flow_datastore.TYPE,
            "METAFLOW_DEFAULT_METADATA": "service",
            "METAFLOW_KUBERNETES_WORKLOAD": str(
                1
            ),  # This is used by kubernetes decorator.
            "METAFLOW_RUNTIME_ENVIRONMENT": "kubernetes",
            "METAFLOW_CARD_S3ROOT": CARD_S3ROOT,
            "METAFLOW_RUN_ID": AIRFLOW_MACROS.RUN_ID,
            "METAFLOW_AIRFLOW_TASK_ID": AIRFLOW_MACROS.create_task_id(
                self.contains_foreach
            ),
            "METAFLOW_AIRFLOW_DAG_RUN_ID": AIRFLOW_MACROS.AIRFLOW_RUN_ID,
            "METAFLOW_AIRFLOW_JOB_ID": AIRFLOW_MACROS.AIRFLOW_JOB_ID,
            "METAFLOW_PRODUCTION_TOKEN": self.production_token,
            "METAFLOW_ATTEMPT_NUMBER": AIRFLOW_MACROS.ATTEMPT,
            # GCP stuff
            "METAFLOW_DATASTORE_SYSROOT_GS": DATASTORE_SYSROOT_GS,
            "METAFLOW_CARD_GSROOT": CARD_GSROOT,
            "METAFLOW_S3_ENDPOINT_URL": S3_ENDPOINT_URL,
        }
        env[
            "METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT"
        ] = AZURE_STORAGE_BLOB_SERVICE_ENDPOINT
        env["METAFLOW_DATASTORE_SYSROOT_AZURE"] = DATASTORE_SYSROOT_AZURE
        env["METAFLOW_CARD_AZUREROOT"] = CARD_AZUREROOT
        if DEFAULT_SECRETS_BACKEND_TYPE:
            env["METAFLOW_DEFAULT_SECRETS_BACKEND_TYPE"] = DEFAULT_SECRETS_BACKEND_TYPE
        if AWS_SECRETS_MANAGER_DEFAULT_REGION:
            env[
                "METAFLOW_AWS_SECRETS_MANAGER_DEFAULT_REGION"
            ] = AWS_SECRETS_MANAGER_DEFAULT_REGION

        env.update(additional_mf_variables)

        service_account = (
            KUBERNETES_SERVICE_ACCOUNT
            if k8s_deco.attributes["service_account"] is None
            else k8s_deco.attributes["service_account"]
        )
        k8s_namespace = (
            k8s_deco.attributes["namespace"]
            if k8s_deco.attributes["namespace"] is not None
            else "default"
        )

        resources = dict(
            requests={
                "cpu": k8s_deco.attributes["cpu"],
                "memory": "%sM" % str(k8s_deco.attributes["memory"]),
                "ephemeral-storage": str(k8s_deco.attributes["disk"]),
            }
        )
        if k8s_deco.attributes["gpu"] is not None:
            resources.update(
                dict(
                    limits={
                        "%s.com/gpu".lower()
                        % k8s_deco.attributes["gpu_vendor"]: str(
                            k8s_deco.attributes["gpu"]
                        )
                    }
                )
            )

        annotations = {
            "metaflow/production_token": self.production_token,
            "metaflow/owner": self.username,
            "metaflow/user": self.username,
            "metaflow/flow_name": self.flow.name,
        }
        if current.get("project_name"):
            annotations.update(
                {
                    "metaflow/project_name": current.project_name,
                    "metaflow/branch_name": current.branch_name,
                    "metaflow/project_flow_name": current.project_flow_name,
                }
            )

        k8s_operator_args = dict(
            # like argo workflows we use step_name as name of container
            name=node.name,
            namespace=k8s_namespace,
            service_account_name=service_account,
            node_selector=k8s_deco.attributes["node_selector"],
            cmds=k8s._command(
                self.flow.name,
                AIRFLOW_MACROS.RUN_ID,
                node.name,
                AIRFLOW_MACROS.create_task_id(self.contains_foreach),
                AIRFLOW_MACROS.ATTEMPT,
                code_package_url=self.code_package_url,
                step_cmds=self._step_cli(
                    node, input_paths, self.code_package_url, user_code_retries
                ),
            ),
            annotations=annotations,
            image=k8s_deco.attributes["image"],
            resources=resources,
            execution_timeout=dict(seconds=runtime_limit),
            retries=user_code_retries,
            env_vars=[dict(name=k, value=v) for k, v in env.items() if v is not None],
            labels=labels,
            task_id=node.name,
            startup_timeout_seconds=AIRFLOW_KUBERNETES_STARTUP_TIMEOUT_SECONDS,
            get_logs=True,
            do_xcom_push=True,
            log_events_on_failure=True,
            is_delete_operator_pod=True,
            retry_exponential_backoff=False,  # todo : should this be a arg we allow on CLI. not right now - there is an open ticket for this - maybe at some point we will.
            reattach_on_restart=False,
            secrets=[],
        )
        k8s_operator_args["in_cluster"] = True
        if AIRFLOW_KUBERNETES_CONN_ID is not None:
            k8s_operator_args["kubernetes_conn_id"] = AIRFLOW_KUBERNETES_CONN_ID
            k8s_operator_args["in_cluster"] = False
        if AIRFLOW_KUBERNETES_KUBECONFIG_CONTEXT is not None:
            k8s_operator_args["cluster_context"] = AIRFLOW_KUBERNETES_KUBECONFIG_CONTEXT
            k8s_operator_args["in_cluster"] = False
        if AIRFLOW_KUBERNETES_KUBECONFIG_FILE is not None:
            k8s_operator_args["config_file"] = AIRFLOW_KUBERNETES_KUBECONFIG_FILE
            k8s_operator_args["in_cluster"] = False

        if k8s_deco.attributes["secrets"]:
            if isinstance(k8s_deco.attributes["secrets"], str):
                k8s_operator_args["secrets"] = k8s_deco.attributes["secrets"].split(",")
            elif isinstance(k8s_deco.attributes["secrets"], list):
                k8s_operator_args["secrets"] = k8s_deco.attributes["secrets"]
        if len(KUBERNETES_SECRETS) > 0:
            k8s_operator_args["secrets"] += KUBERNETES_SECRETS.split(",")

        if retry_delay:
            k8s_operator_args["retry_delay"] = dict(seconds=retry_delay.total_seconds())

        return k8s_operator_args

    def _step_cli(self, node, paths, code_package_url, user_code_retries):
        cmds = []

        script_name = os.path.basename(sys.argv[0])
        executable = self.environment.executable(node.name)

        entrypoint = [executable, script_name]

        top_opts_dict = {
            "with": [
                decorator.make_decorator_spec()
                for decorator in node.decorators
                if not decorator.statically_defined
            ]
        }
        # FlowDecorators can define their own top-level options. They are
        # responsible for adding their own top-level options and values through
        # the get_top_level_options() hook. See similar logic in runtime.py.
        for deco in flow_decorators():
            top_opts_dict.update(deco.get_top_level_options())

        top_opts = list(dict_to_cli_options(top_opts_dict))

        top_level = top_opts + [
            "--quiet",
            "--metadata=%s" % self.metadata.TYPE,
            "--environment=%s" % self.environment.TYPE,
            "--datastore=%s" % self.flow_datastore.TYPE,
            "--datastore-root=%s" % self.flow_datastore.datastore_root,
            "--event-logger=%s" % self.event_logger.TYPE,
            "--monitor=%s" % self.monitor.TYPE,
            "--no-pylint",
            "--with=airflow_internal",
        ]

        if node.name == "start":
            # We need a separate unique ID for the special _parameters task
            task_id_params = "%s-params" % AIRFLOW_MACROS.create_task_id(
                self.contains_foreach
            )
            # Export user-defined parameters into runtime environment
            param_file = "".join(
                random.choice(string.ascii_lowercase) for _ in range(10)
            )
            # Setup Parameters as environment variables which are stored in a dictionary.
            export_params = (
                "python -m "
                "metaflow.plugins.airflow.plumbing.set_parameters %s "
                "&& . `pwd`/%s" % (param_file, param_file)
            )
            # Setting parameters over here.
            params = (
                entrypoint
                + top_level
                + [
                    "init",
                    "--run-id %s" % AIRFLOW_MACROS.RUN_ID,
                    "--task-id %s" % task_id_params,
                ]
            )

            # Assign tags to run objects.
            if self.tags:
                params.extend("--tag %s" % tag for tag in self.tags)

            # If the start step gets retried, we must be careful not to
            # regenerate multiple parameters tasks. Hence, we check first if
            # _parameters exists already.
            exists = entrypoint + [
                # Dump the parameters task
                "dump",
                "--max-value-size=0",
                "%s/_parameters/%s" % (AIRFLOW_MACROS.RUN_ID, task_id_params),
            ]
            cmd = "if ! %s >/dev/null 2>/dev/null; then %s && %s; fi" % (
                " ".join(exists),
                export_params,
                " ".join(params),
            )
            cmds.append(cmd)
            # set input paths for parameters
            paths = "%s/_parameters/%s" % (AIRFLOW_MACROS.RUN_ID, task_id_params)

        step = [
            "step",
            node.name,
            "--run-id %s" % AIRFLOW_MACROS.RUN_ID,
            "--task-id %s" % AIRFLOW_MACROS.create_task_id(self.contains_foreach),
            "--retry-count %s" % AIRFLOW_MACROS.ATTEMPT,
            "--max-user-code-retries %d" % user_code_retries,
            "--input-paths %s" % paths,
        ]
        if self.tags:
            step.extend("--tag %s" % tag for tag in self.tags)
        if self.namespace is not None:
            step.append("--namespace=%s" % self.namespace)

        parent_is_foreach = any(  # The immediate parent is a foreach node.
            self.graph[n].type == "foreach" for n in node.in_funcs
        )
        if parent_is_foreach:
            step.append("--split-index %s" % AIRFLOW_MACROS.FOREACH_SPLIT_INDEX)

        cmds.append(" ".join(entrypoint + top_level + step))
        return cmds

    def _collect_flow_sensors(self):
        decos_lists = [
            self.flow._flow_decorators.get(s.name)
            for s in SUPPORTED_SENSORS
            if self.flow._flow_decorators.get(s.name) is not None
        ]
        af_tasks = [deco.create_task() for decos in decos_lists for deco in decos]
        if len(af_tasks) > 0:
            self._depends_on_upstream_sensors = True
        return af_tasks

    def _contains_foreach(self):
        for node in self.graph:
            if node.type == "foreach":
                return True
        return False

    def compile(self):
        if self.flow._flow_decorators.get("trigger") or self.flow._flow_decorators.get(
            "trigger_on_finish"
        ):
            raise AirflowException(
                "Deploying flows with @trigger or @trigger_on_finish decorator(s) "
                "to Airflow is not supported currently."
            )

        # Visit every node of the flow and recursively build the state machine.
        def _visit(node, workflow, exit_node=None):
            kube_deco = dict(
                [deco for deco in node.decorators if deco.name == "kubernetes"][
                    0
                ].attributes
            )
            if kube_deco:
                # Only guard against use_tmpfs and tmpfs_size as these determine if tmpfs is enabled.
                for attr in [
                    "use_tmpfs",
                    "tmpfs_size",
                    "persistent_volume_claims",
                    "image_pull_policy",
                ]:
                    if kube_deco[attr]:
                        raise AirflowException(
                            "The decorator attribute *%s* is currently not supported on Airflow "
                            "for the @kubernetes decorator on step *%s*"
                            % (attr, node.name)
                        )

            parent_is_foreach = any(  # Any immediate parent is a foreach node.
                self.graph[n].type == "foreach" for n in node.in_funcs
            )
            state = AirflowTask(
                node.name, is_mapper_node=parent_is_foreach
            ).set_operator_args(**self._to_job(node))
            if node.type == "end":
                workflow.add_state(state)

            # Continue linear assignment within the (sub)workflow if the node
            # doesn't branch or fork.
            elif node.type in ("start", "linear", "join", "foreach"):
                workflow.add_state(state)
                _visit(
                    self.graph[node.out_funcs[0]],
                    workflow,
                )

            elif node.type == "split":
                workflow.add_state(state)
                for func in node.out_funcs:
                    _visit(
                        self.graph[func],
                        workflow,
                    )
            else:
                raise AirflowException(
                    "Node type *%s* for  step *%s* "
                    "is not currently supported by "
                    "Airflow." % (node.type, node.name)
                )

            return workflow

        # set max active tasks here , For more info check here :
        # https://airflow.apache.org/docs/apache-airflow/stable/_api/airflow/models/dag/index.html#airflow.models.dag.DAG
        airflow_dag_args = (
            {} if self.max_workers is None else dict(max_active_tasks=self.max_workers)
        )
        airflow_dag_args["is_paused_upon_creation"] = self.is_paused_upon_creation

        # workflow timeout should only be enforced if a dag is scheduled.
        if self.workflow_timeout is not None and self.schedule is not None:
            airflow_dag_args["dagrun_timeout"] = dict(seconds=self.workflow_timeout)

        appending_sensors = self._collect_flow_sensors()
        workflow = Workflow(
            dag_id=self.name,
            default_args=self._create_defaults(),
            description=self.description,
            schedule_interval=self.schedule,
            # `start_date` is a mandatory argument even though the documentation lists it as optional value
            # Based on the code, Airflow will throw a `AirflowException` when `start_date` is not provided
            # to a DAG : https://github.com/apache/airflow/blob/0527a0b6ce506434a23bc2a6f5ddb11f492fc614/airflow/models/dag.py#L2170
            start_date=datetime.now(),
            tags=self.tags,
            file_path=self._file_path,
            graph_structure=self.graph_structure,
            metadata=dict(
                contains_foreach=self.contains_foreach, flow_name=self.flow.name
            ),
            **airflow_dag_args
        )
        workflow = _visit(self.graph["start"], workflow)

        workflow.set_parameters(self.parameters)
        if len(appending_sensors) > 0:
            for s in appending_sensors:
                workflow.add_state(s)
            workflow.graph_structure.insert(0, [[s.name] for s in appending_sensors])
        return self._to_airflow_dag_file(workflow.to_dict())

    def _to_airflow_dag_file(self, json_dag):
        util_file = None
        with open(airflow_utils.__file__) as f:
            util_file = f.read()
        with open(AIRFLOW_DEPLOY_TEMPLATE_FILE) as f:
            return chevron.render(
                f.read(),
                dict(
                    # Converting the configuration to base64 so that there can be no indentation related issues that can be caused because of
                    # malformed strings / json.
                    config=json_dag,
                    utils=util_file,
                    deployed_on=str(datetime.now()),
                ),
            )

    def _create_defaults(self):
        defu_ = {
            "owner": get_username(),
            # If set on a task and the previous run of the task has failed,
            # it will not run the task in the current DAG run.
            "depends_on_past": False,
            # TODO: Enable emails
            "execution_timeout": timedelta(days=5),
            "retry_delay": timedelta(seconds=200),
            # check https://airflow.apache.org/docs/apache-airflow/stable/_api/airflow/models/baseoperator/index.html?highlight=retry_delay#airflow.models.baseoperator.BaseOperatorMeta
        }
        if self.worker_pool is not None:
            defu_["pool"] = self.worker_pool

        return defu_
